# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

import glob
import os
import sys

import bleachbit
from bleachbit import APP_NAME, Cleaner, GuiBasic, appicon_path, portable_mode
from bleachbit.Cleaner import backends
from bleachbit.GtkShim import (
    GLib, Gdk, Gio, Gtk,
    require_gtk,
    suppress_pygobject_asyncio_warnings,
)
from bleachbit.GUI import logger
from bleachbit.GuiUtil import get_clipboard_paths
from bleachbit.GuiWindow import GUI
from bleachbit.Language import get_text as _, get_active_language_code
from bleachbit.Options import options

# Ensure GTK is available for this GUI module
require_gtk()

# Constants for dialog response codes
RESPONSE_ANONYMIZE = 101
RESPONSE_COPY = 100

if os.name == 'nt':
    from bleachbit import Windows
    from bleachbit.FontCheckDialog import (
        create_font_check_dialog,
        RESPONSE_TEXT_BLURRY,
        RESPONSE_TEXT_UNREADABLE,
    )


class Bleachbit(Gtk.Application):
    _window = None
    _shred_paths = None
    _auto_exit = False

    def __init__(self, uac=True, shred_paths=None, auto_exit=False):

        application_id_suffix = self._init_windows_misc(
            auto_exit, shred_paths, uac)
        # Support pytest-xdist parallel workers by making application ID unique
        xdist_worker = os.environ.get('PYTEST_XDIST_WORKER', '')
        if xdist_worker:
            application_id_suffix += xdist_worker
        application_id = '{}{}'.format(
            'org.gnome.Bleachbit', application_id_suffix)
        Gtk.Application.__init__(
            self, application_id=application_id, flags=Gio.ApplicationFlags.FLAGS_NONE)
        GLib.set_prgname('org.bleachbit.BleachBit')

        self._font_check_prompt_scheduled = False

        if auto_exit:
            # This is used for automated testing of whether the GUI can start.
            # It is called from assert_execute_console() in windows/setup.py
            self._auto_exit = True

        if shred_paths:
            self._shred_paths = shred_paths

        if os.name == 'nt':
            # clean up nonce files https://github.com/bleachbit/bleachbit/issues/858
            import atexit
            atexit.register(Windows.cleanup_nonce)

    def run(self, *args, **kwargs):
        """Run the GTK application."""
        with suppress_pygobject_asyncio_warnings():
            return Gtk.Application.run(self, *args, **kwargs)

    def _init_windows_misc(self, auto_exit, shred_paths, uac):
        application_id_suffix = ''
        is_context_menu_executed = auto_exit and shred_paths
        if not os.name == 'nt':
            return ''
        env_suffix = os.environ.pop('BLEACHBIT_APP_INSTANCE_SUFFIX', '')
        if env_suffix:
            application_id_suffix = env_suffix
        if Windows.elevate_privileges(uac):
            # privileges escalated in other process
            sys.exit(0)

        if is_context_menu_executed:
            # When we have a running application and executing the Windows
            # context menu command we start a new process with new application_id.
            # That is because the command line arguments of the context menu command
            # are not passed to the already running instance.
            application_id_suffix = 'ContextMenuShred'
        return application_id_suffix

    def build_app_menu(self):
        """Build the application menu

        On Linux with GTK 3.24, this code is necessary but not sufficient for
        the menu to work. The headerbar code is also needed.

        On Windows with GTK 3.18, this code is sufficient for the menu to work.
        """
        from bleachbit.Language import setup_translation
        setup_translation()

        # set_translation_domain() seems to have no effect.
        # builder.set_translation_domain('bleachbit')
        app_menu_path = bleachbit.get_share_path('app-menu.ui')
        if app_menu_path:
            builder = Gtk.Builder()
            builder.add_from_file(app_menu_path)
            menu = builder.get_object('app-menu')
            self.set_app_menu(menu)
        else:
            logger.error('build_app_menu(): app-menu.ui not found')

        # set up mappings between <attribute name="action"> in app-menu.ui and methods in this class
        actions = {'shredFiles': self.cb_shred_file,
                   'shredFolders': self.cb_shred_folder,
                   'shredClipboard': self.cb_shred_clipboard,
                   'wipeEmptySpace': self.cb_wipe_empty_space,
                   'makeChaff': self.cb_make_chaff,
                   'shredQuit': self.cb_shred_quit,
                   'preferences': self.cb_preferences_dialog,
                   'systemInformation': self.system_information_dialog,
                   'help': self.cb_help,
                   'about': self.about}

        for action_name, callback in actions.items():
            action = Gio.SimpleAction.new(action_name, None)
            action.connect('activate', callback)
            self.add_action(action)

    def cb_help(self, action, param):
        """Callback for help"""
        GuiBasic.open_url(bleachbit.help_contents_url, self._window)

    def cb_make_chaff(self, action, param):
        """Callback to make chaff"""
        from bleachbit.GuiChaff import ChaffDialog
        cd = ChaffDialog(self._window)
        cd.run()

    def cb_shred_file(self, action, param):
        """Callback for shredding a file"""

        # get list of files
        # TRANSLATORS: Title of a file chooser dialog.
        paths = GuiBasic.browse_files(self._window, _("Choose files to shred"))
        if not paths:
            return
        GUI.shred_paths(self._window, paths)

    def cb_shred_folder(self, action, param):
        """Callback for shredding a folder"""

        # TRANSLATORS: Title of a folder chooser dialog.
        title = _("Choose folder to shred")
        # TRANSLATORS: Button label in a folder chooser dialog.
        button_label = _('_Delete')
        paths = GuiBasic.browse_folder(self._window,
                                       title,
                                       multiple=True,
                                       stock_button=button_label)
        if not paths:
            return
        GUI.shred_paths(self._window, paths)

    def cb_shred_clipboard(self, action, param):
        """Callback for menu option: shred paths from clipboard"""
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        # wait_for_targets() avoids GLib warnings from request_targets() when
        # the clipboard is empty.
        has_targets, targets = clipboard.wait_for_targets()
        if not has_targets:
            targets = []
        self.cb_clipboard_uri_received(clipboard, targets, None)

    def cb_clipboard_uri_received(self, clipboard, targets, _data):
        """Callback for when URIs are received from clipboard

        With GTK 3.18.9 on Windows, there was no text/uri-list in targets,
        but there is with GTK 3.24.34. However, Windows does not have
        get_uris().
        """
        shred_paths = get_clipboard_paths(clipboard, targets)
        # TRANSLATORS: Warning log message when attempting to paste files/folders to shred.
        not_found_msg = _('No paths found in clipboard.')
        if shred_paths:
            GUI.shred_paths(self._window, shred_paths,
                            should_clear_clipboard=True)
        else:
            logger.warning(not_found_msg)

    def cb_shred_quit(self, action, param):
        """Shred settings (for privacy reasons) and quit"""
        # build a list of paths to delete
        paths = []
        if os.name == 'nt' and portable_mode:
            # in portable mode on Windows, the options directory includes
            # executables
            paths.append(bleachbit.options_file)
            if os.path.isdir(bleachbit.personal_cleaners_dir):
                paths.append(bleachbit.personal_cleaners_dir)
            for f in glob.glob(os.path.join(bleachbit.options_dir, "*.bz2")):
                paths.append(f)
        else:
            paths.append(bleachbit.options_dir)

        # prompt the user to confirm
        if not GUI.shred_paths(self._window, paths, shred_settings=True):
            logger.debug('user aborted shred')
            # aborted
            return

        # Quit the application through the idle loop to allow the worker
        # to delete the files.  Use the lowest priority because the worker
        # uses the standard priority. Otherwise, this will quit before
        # the files are deleted.
        #
        # Rebuild a minimal bleachbit.ini when quitting
        GLib.idle_add(self.quit, None, None, True,
                      priority=GLib.PRIORITY_LOW)

    def cb_wipe_empty_space(self, action, param):
        """callback to wipe empty space in arbitrary folder"""
        # TRANSLATORS: Title of a folder chooser dialog.
        title = _("Choose a folder")
        # TRANSLATORS: Button label in a folder chooser dialog.
        # Underscore is for accelerator key.
        button_label = _('_OK')
        path = GuiBasic.browse_folder(self._window,
                                      title,
                                      multiple=False,
                                      stock_button=button_label)
        if not path:
            # user cancelled
            return

        backends['_gui'] = Cleaner.create_wipe_empty_space_cleaner(path)

        # execute
        operations = {'_gui': ['empty_space']}
        self._window.preview_or_run_operations(True, operations)

    def get_preferences_dialog(self):
        return self._window.get_preferences_dialog()

    def cb_preferences_dialog(self, action, param):
        """Callback for preferences dialog"""
        self._window.show_preferences_dialog()

    def get_about_dialog(self):
        # TRANSLATORS: Title of the 'About' dialog.
        dialog = Gtk.AboutDialog(comments=_("Program to clean unnecessary files"),
                                 copyright=bleachbit.APP_COPYRIGHT,
                                 program_name=APP_NAME,
                                 version=bleachbit.APP_VERSION,
                                 website=bleachbit.APP_URL,
                                 transient_for=self._window)
        try:
            with open(bleachbit.license_filename) as f_license:
                dialog.set_license(f_license.read())
        except (IOError, TypeError):
            # TRANSLATORS: License text shown in the 'About' dialog.
            license_msg = _("GNU General Public License version 3 or later.\n"
                            "See https://www.gnu.org/licenses/gpl-3.0.txt")
            dialog.set_license(license_msg)
        # TRANSLATORS: Maintain the names of translators here.
        # Launchpad does this automatically for translations
        # typed in Launchpad. This is a special string shown
        # in the 'About' box.
        dialog.set_translator_credits(_("translator-credits"))
        if appicon_path and os.path.exists(appicon_path):
            icon = Gtk.Image.new_from_file(appicon_path)
            dialog.set_logo(icon.get_pixbuf())

        return dialog

    def about(self, _action, _param):
        """Create and show the about dialog"""
        dialog = self.get_about_dialog()
        dialog.run()
        dialog.destroy()

    def do_startup(self):
        Gtk.Application.do_startup(self)
        self.build_app_menu()

    def quit(self, _action=None, _param=None, init_configuration=False):
        if init_configuration:
            # After "shred settings and quit", rebuild the minimal configuration.
            # which is important for portable mode.
            bleachbit.Options.init_configuration()
        self._window.destroy()

    def get_system_information_dialog(self):
        """Show system information dialog"""
        # TRANSLATORS: Title of the system information dialog.
        dialog = Gtk.Dialog(title=_("System information"),
                            transient_for=self._window)
        dialog.set_default_size(600, 400)
        txtbuffer = Gtk.TextBuffer()
        from bleachbit.SystemInformation import get_system_information
        txt = get_system_information()
        txtbuffer.set_text(txt)
        textview = Gtk.TextView.new_with_buffer(txtbuffer)
        textview.set_editable(False)
        swindow = Gtk.ScrolledWindow()
        swindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        swindow.add(textview)
        dialog.get_content_area().pack_start(swindow, True, True, 0)
        # TRANSLATORS: Button label in the system information dialog to
        # replace the username with a placeholder.
        dialog.add_buttons(_("Anonymize"), RESPONSE_ANONYMIZE,
                           Gtk.STOCK_COPY, RESPONSE_COPY,
                           Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
        return (dialog, txt, txtbuffer)

    def system_information_dialog(self, _action, _param):
        from bleachbit.SystemInformation import anonymize_system_information
        dialog, txt, txtbuffer = self.get_system_information_dialog()
        dialog.show_all()
        while True:
            rc = dialog.run()
            if rc == RESPONSE_COPY:
                clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
                current_text = txtbuffer.get_text(
                    txtbuffer.get_start_iter(),
                    txtbuffer.get_end_iter(),
                    True)
                clipboard.set_text(current_text, -1)
            elif rc == RESPONSE_ANONYMIZE:
                anonymized_txt = anonymize_system_information(txt)
                txtbuffer.set_text(anonymized_txt)
                # The button is single use.
                dialog.get_widget_for_response(
                    RESPONSE_ANONYMIZE).set_sensitive(False)
            else:
                break
        dialog.destroy()

    def do_activate(self):
        if not self._window:
            self._window = GUI(
                application=self, title=APP_NAME, auto_exit=self._auto_exit)
        self._window.present()
        if self._shred_paths:
            GLib.idle_add(GUI.shred_paths, self._window,
                          self._shred_paths, priority=GLib.PRIORITY_LOW)
            # When we shred paths and auto exit with the Windows Explorer context menu command we close the
            # application in GUI.shred_paths, because if it is closed from here there are problems.
            # Most probably this is something related with how GTK handles idle quit calls.
        elif self._auto_exit:
            GLib.idle_add(self.quit,
                          priority=GLib.PRIORITY_LOW)
            print('Success')
        else:
            # Check for orphaned wipe files from interrupted operations
            GLib.idle_add(self._window.check_orphaned_wipe_files,
                          priority=GLib.PRIORITY_LOW)

        self._maybe_prompt_font_check()

    def _should_show_font_check_dialog(self):
        """Determine whether to show the font check dialog on Windows."""
        if os.name != 'nt':
            return False
        if self._auto_exit:
            return False
        if self._shred_paths:
            return False
        # User made an explicit choice of a backend.
        if os.environ.get('PANGOCAIRO_BACKEND', ''):
            return False
        if options.get('use_fontconfig_backend'):
            return False
        if options.get('font_check_completed'):
            return False
        # Skip for CJK languages (Chinese, Japanese, Korean).
        # They do not support Arial font, and they seem not to
        # be affected by the font bug.
        lang = get_active_language_code()
        if lang.startswith(('zh', 'ja', 'ko')):
            return False
        return True

    def _maybe_prompt_font_check(self):
        """Schedule the font check dialog if needed."""
        if not self._should_show_font_check_dialog():
            return
        if self._font_check_prompt_scheduled:
            return
        self._font_check_prompt_scheduled = True
        GLib.idle_add(self._show_font_check_dialog,
                      priority=GLib.PRIORITY_DEFAULT_IDLE)

    def _show_font_check_dialog(self):
        """Show the font check dialog and handle the response."""
        if os.name != 'nt':
            return False
        if not self._window:
            return False
        dialog = create_font_check_dialog(self._window)
        response = dialog.run()
        dialog.destroy()
        if response == Gtk.ResponseType.YES:
            options.set('font_check_completed', True)
        elif response in (RESPONSE_TEXT_BLURRY, RESPONSE_TEXT_UNREADABLE):
            self._restart_with_fontconfig_backend()
        # If user dismisses the dialog, ask again next time.
        return False

    def _restart_with_fontconfig_backend(self):
        """Restart BleachBit with the fontconfig Pango backend."""
        from bleachbit import General
        executable = General.get_executable()
        if getattr(sys, 'frozen', False):
            cmd = [executable] + sys.argv[1:]
        else:
            script_path = os.path.abspath(sys.argv[0])
            cmd = [executable, script_path] + sys.argv[1:]
        logger.info('Restarting BleachBit with fontconfig backend: %s', cmd)
        env = os.environ.copy()
        env['PANGOCAIRO_BACKEND'] = 'fc'
        env['BLEACHBIT_APP_INSTANCE_SUFFIX'] = f'Restart{os.getpid()}'
        options.set('font_check_completed', True)
        options.set('use_fontconfig_backend', True)
        if General.run_external_nowait(cmd, env=env):
            self.quit()
        else:
            # This logs to the main application window
            logger.error('Failed to restart BleachBit with fontconfig backend')
            options.set('font_check_completed', False)
            options.set('use_fontconfig_backend', False)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

from bleachbit import GuiBasic
from bleachbit.Cleaner import backends
from bleachbit.GUI import logger
from bleachbit.GtkShim import GObject, Gtk
from bleachbit.Language import get_text as _
from bleachbit.Options import options


class TreeDisplayModel:
    """Displays the info model in a view"""

    def make_view(self, model, parent, context_menu_event):
        """Create and return a TreeView object"""
        self.view = Gtk.TreeView.new_with_model(model)

        # hide headers
        # These commits disabled the header: 5d0b5b7ab,49ad443.
        self.view.set_headers_visible(False)

        # listen for right click (context menu)
        self.view.connect("button_press_event", context_menu_event)

        # first column: cleaner name
        renderer_name = Gtk.CellRendererText()
        # TRANSLATORS: Column header for the cleaner name.
        # Used in the tree view on the main screen.
        # 'Name' is a noun.
        column_name = Gtk.TreeViewColumn(_("Name"), renderer_name, text=0)
        self.view.append_column(column_name)
        self.view.set_search_column(0)

        # second column: alert icon
        renderer_alert = Gtk.CellRendererPixbuf()
        column_alert = Gtk.TreeViewColumn("", renderer_alert)
        column_alert.add_attribute(renderer_alert, "icon-name", 4)
        self.view.append_column(column_alert)

        # third column: checkbox
        renderer_active = Gtk.CellRendererToggle()
        renderer_active.set_property('activatable', True)
        renderer_active.connect('toggled', self.col1_toggled_cb, model, parent)
        # TRANSLATORS: Column header indicating whether the cleaner is active/enabled.
        # Used in the tree view on the main screen.
        column_active = Gtk.TreeViewColumn(_("Active"), renderer_active)
        column_active.add_attribute(renderer_active, "active", 1)
        self.view.append_column(column_active)

        # fourth column: size
        renderer_size = Gtk.CellRendererText()
        renderer_size.set_alignment(1.0, 0.0)
        # TRANSLATORS: Column header for the size of the cleaner.
        # Used in the tree view on the main screen.
        # 'Size' is a noun and refers to the amount of space that the cleaner
        # would clean (preview mode) or did clean (cleaning mode).
        column_size = Gtk.TreeViewColumn(_("Size"), renderer_size, text=3)
        column_size.set_alignment(1.0)
        self.view.append_column(column_size)

        # finish
        self.view.expand_all()
        return self.view

    def set_cleaner(self, path, model, parent_window, value):
        """Activate or deactivate option of cleaner."""
        assert isinstance(value, bool)
        assert isinstance(model, Gtk.TreeStore)
        cleaner_id = None
        i = path
        if isinstance(i, str):
            # type is either str or gtk.TreeIter
            i = model.get_iter(path)
        parent = model.iter_parent(i)
        if parent:
            # this is an option (child), not a cleaner (parent)
            cleaner_id = model[parent][2]
            option_id = model[path][2]
        if cleaner_id and value:
            # When enabling an option, present any warnings.
            # (When disabling an option, there is no need to present warnings.)
            warning = backends[cleaner_id].get_warning(option_id)
            if warning and not options.get('expert_mode'):
                if hasattr(parent_window, 'show_infobar'):
                    parent_window.show_infobar(
                        # TRANSLATORS: Warning shown in the infobar.
                        _("This option is protected. To bypass protection, enable expert mode."))
                # Show an alert icon by the option name.
                model[path][4] = "dialog-warning"
                return
            warning_key = 'cleaner:' + cleaner_id + ':' + option_id
            if warning and not options.get_warning_preference(warning_key):
                # TRANSLATORS: %(cleaner) may be Firefox, System, etc.
                # %(option) may be cache, logs, cookies, etc.
                option_name = _("%(cleaner)s - %(option)s") % {
                    'cleaner': model[parent][0],
                    'option': model[path][0],
                }
                confirmed, remember_choice = GuiBasic.warning_confirm_dialog(
                    parent_window, option_name, warning)
                if not confirmed:
                    # user cancelled, so don't toggle option
                    return
                if remember_choice:
                    options.remember_warning_preference(warning_key)
        model[path][1] = value
        model[path][4] = ""

    def col1_toggled_cb(self, cell, path, model, parent_window):
        """Callback for toggling cleaners"""
        is_toggled_on = not model[path][1]  # Is the new state enabled?
        self.set_cleaner(path, model, parent_window, is_toggled_on)
        i = model.get_iter(path)
        parent = model.iter_parent(i)
        if parent and is_toggled_on:
            # If child is enabled, then also enable the parent.
            model[parent][1] = True
        # If all siblings were toggled off, then also disable the parent.
        if parent and not is_toggled_on:
            sibling = model.iter_nth_child(parent, 0)
            any_sibling_enabled = False
            while sibling:
                if model[sibling][1]:
                    any_sibling_enabled = True
                sibling = model.iter_next(sibling)
            if not any_sibling_enabled:
                model[parent][1] = False
        # If toggled and has children, then do the same for each child.
        child = model.iter_children(i)
        while child:
            self.set_cleaner(child, model, parent_window, is_toggled_on)
            child = model.iter_next(child)
        # If the parent was just enabled but all children were blocked
        # by expert mode, leave the parent unchecked.
        if not parent and is_toggled_on:
            child = model.iter_children(i)
            any_child_enabled = False
            while child:
                if model[child][1]:
                    any_child_enabled = True
                child = model.iter_next(child)
            if not any_child_enabled:
                model[i][1] = False
        return


class TreeInfoModel:
    """Model holds information to be displayed in the tree view"""

    def __init__(self):
        self.tree_store = Gtk.TreeStore(
            GObject.TYPE_STRING, GObject.TYPE_BOOLEAN, GObject.TYPE_PYOBJECT, GObject.TYPE_STRING, GObject.TYPE_STRING)
        if not self.tree_store:
            raise Exception("cannot create tree store")
        self.row_changed_handler_id = None
        self.refresh_rows()
        self.tree_store.set_sort_func(3, self.sort_func)
        self.tree_store.set_sort_column_id(3, Gtk.SortType.ASCENDING)

    def get_model(self):
        """Return the tree store"""
        return self.tree_store

    def on_row_changed(self, __treemodel, path, __iter):
        """Event handler for when a row changes"""
        parent = self.tree_store[path[0]][2]
        child = None
        if len(path) == 2:
            child = self.tree_store[path][2]
        value = self.tree_store[path][1]
        options.set_tree(parent, child, value)

    def refresh_rows(self):
        """Clear rows (cleaners) and add them fresh"""
        if self.row_changed_handler_id:
            self.tree_store.disconnect(self.row_changed_handler_id)
        self.tree_store.clear()
        hidden_cleaners = []
        for key in sorted(backends):
            if not any(backends[key].get_options()):
                # localizations has no options, so it should be hidden
                # https://github.com/az0/bleachbit/issues/110
                continue
            c_name = backends[key].get_name()
            c_id = backends[key].get_id()
            c_value = options.get_tree(c_id, None)
            if not c_value and options.get('auto_hide') and backends[key].auto_hide():
                hidden_cleaners.append(c_id)
                continue
            parent = self.tree_store.append(
                None, (c_name, c_value, c_id, "", ""))
            for (o_id, o_name) in backends[key].get_options():
                o_value = options.get_tree(c_id, o_id)
                if not isinstance(o_name, str):
                    logger.error("Invalid option name type for %s.%s: expected str, got %s: %s",
                                 c_id, o_id, type(o_name).__name__, o_name)
                    continue
                if not isinstance(o_id, str):
                    logger.error("Invalid option id type for %s: expected str, got %s: %s",
                                 c_id, type(o_id).__name__, o_id)
                    continue
                self.tree_store.append(parent, (o_name, o_value, o_id, "", ""))
        if hidden_cleaners:
            logger.debug("automatically hid %d cleaners: %s", len(
                hidden_cleaners), ', '.join(hidden_cleaners))
        self.row_changed_handler_id = self.tree_store.connect("row-changed",
                                                              self.on_row_changed)

    def sort_func(self, model, iter1, iter2, _user_data):
        """Sort the tree by the id

        Index 0 is the display name
        Index 2 is the ID (e.g., cookies, vacuum).

        Sorting by ID is functionally important, so that vacuuming is done
        last, even for other languages. See https://github.com/bleachbit/bleachbit/issues/441
        """
        value1 = model[iter1][2].lower()
        value2 = model[iter2][2].lower()
        if value1 == value2:
            return 0
        if value1 > value2:
            return 1
        return -1



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Prepare to start the application
"""

import getpass
import os
import re
import sys

from bleachbit import IS_POSIX, IS_WINDOWS

# pylint: disable=invalid-name
_bootstrapped = False


def _apply_fontconfig_backend_preference():
    """On Windows, set PANGOCAIRO_BACKEND=fc if the user chose fontconfig.

    It is important that this runs before Gtk.
    """
    if not IS_WINDOWS:
        return
    try:
        from bleachbit.Options import options  # pylint: disable=import-outside-toplevel
        if options.get('use_fontconfig_backend'):
            os.environ['PANGOCAIRO_BACKEND'] = 'fc'
    except Exception:
        pass


def check_wayland_and_root():
    """Check if Wayland is being used and root is not allowed.

    Returns True if there is a problem
    False if no problem (e.g., not root, not Wayland)
    """
    if not IS_POSIX:
        return False

    # The two imports from bleachbit must come after sys.path is adjusted.
    import bleachbit.Unix  # pylint: disable=import-outside-toplevel
    from bleachbit.Language import get_text as _  # pylint: disable=import-outside-toplevel

    # FIXME: if started from launcher (.desktop file), there may be no console
    # to which to print this message.
    if bleachbit.Unix.is_display_protocol_wayland_and_root_not_allowed():
        print(_('To run a GUI application on Wayland with root, allow access with this command:\n'
              'xhost si:localuser:root\n'
                'See more about xhost at https://docs.bleachbit.org/doc/frequently-asked-questions.html'), file=sys.stderr)
        return True
    return False


def _bootstrap_posix():
    """Bootstrap for POSIX systems"""
    # os.path.expanduser('~') returns '~' unchanged when HOME is unset
    # and the user has no passwd entry (e.g., Docker containers).
    from bleachbit import _home_dir, logger
    home_dir = _home_dir()
    if not os.getenv('HOME') and home_dir == '/tmp':
        logger.warning('HOME not set and no passwd entry; using %s', home_dir)

    # Set fallbacks for environment variables.
    envs = {
        'HOME': home_dir,
        'PATH': '/usr/bin:/bin:/usr/sbin:/sbin',
        'XDG_CACHE_HOME': os.path.join(home_dir, '.cache'),
        'XDG_CONFIG_HOME': os.path.join(home_dir, '.config'),
        'XDG_DATA_HOME': os.path.join(home_dir, '.local', 'share')
    }
    if not os.getenv('USER'):
        try:
            envs['USER'] = getpass.getuser()
        except (OSError, KeyError):
            pass
    for varname, value in envs.items():
        if not os.getenv(varname):
            os.environ[varname] = value


def _bootstrap_windows():
    """Bootstrap for Windows"""
    from bleachbit import Windows
    Windows.setup_environment()

    # Use our `font.conf` (see commit 3385952b37d78).
    os.environ.pop('FONTCONFIG_FILE', None)

    # change error handling to avoid popup with GTK 3
    # https://github.com/bleachbit/bleachbit/issues/651
    import win32api  # pylint: disable=import-error
    import win32con  # pylint: disable=import-error
    win32api.SetErrorMode(win32con.SEM_FAILCRITICALERRORS |
                          win32con.SEM_NOGPFAULTERRORBOX | win32con.SEM_NOOPENFILEERRORBOX)

    # Set GDK_PIXBUF_MODULE_FILE based on the Python DLL location.
    # This ensures GTK can find the pixbuf loaders when running from
    # a bundled/frozen environment where the standard paths may not apply.
    import win32process
    for process in win32process.EnumProcessModules(-1):
        name = win32process.GetModuleFileNameEx(-1, process)
        if re.search(r'python\d+.dll$', name, re.IGNORECASE):
            bindir = os.path.dirname(name)
            os.environ['GDK_PIXBUF_MODULE_FILE'] = os.path.join(
                bindir, 'lib', 'gdk-pixbuf-2.0', '2.10.0', 'loaders.cache')


def bootstrap():
    """Bootstrap the application"""
    global _bootstrapped  # pylint: disable=global-statement
    if _bootstrapped:
        return
    _bootstrapped = True
    _apply_fontconfig_backend_preference()
    if IS_WINDOWS:
        _bootstrap_windows()
    elif IS_POSIX:
        _bootstrap_posix()



# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Basic GUI code
"""

# standard library
import os

# local import
from bleachbit.GtkShim import Gtk, Gdk, GLib, require_gtk
from bleachbit.Language import get_text as _
from bleachbit.Options import options
if os.name == 'nt':
    from bleachbit import Windows

# Ensure GTK is available for this GUI module
require_gtk()


# TRANSLATORS: Label for the Cancel button in several dialog windows:
# file chooser, folder choose, delete confirmation, and warning dialogs.
# The underscore indicates the accelerator key.
CANCEL_BUTTON_LABEL = _("_Cancel")
# TRANSLATORS: Label for the Delete button in the file choose dialog
# and delete confirmation dialog.
DELETE_BUTTON_LABEL = _("_Delete")


def browse_folder(parent, title, multiple, stock_button):
    """Ask the user to select a folder.  Return the full path or None."""

    if os.name == 'nt' and not os.getenv('BB_NATIVE'):
        ret = Windows.browse_folder(parent, title)
        return [ret] if multiple and not ret is None else ret

    # fall back to GTK+
    chooser = Gtk.FileChooserDialog(transient_for=parent,
                                    title=title,
                                    action=Gtk.FileChooserAction.SELECT_FOLDER)
    chooser.add_buttons(CANCEL_BUTTON_LABEL, Gtk.ResponseType.CANCEL,
                        stock_button, Gtk.ResponseType.OK)
    chooser.set_default_response(Gtk.ResponseType.OK)
    chooser.set_select_multiple(multiple)
    chooser.set_current_folder(os.path.expanduser('~'))
    resp = chooser.run()
    if multiple:
        ret = chooser.get_filenames()
    else:
        ret = chooser.get_filename()
    chooser.hide()
    chooser.destroy()
    if Gtk.ResponseType.OK != resp:
        # user cancelled
        return None
    return ret


def browse_file(parent, title):
    """Prompt user to select a single file"""

    if os.name == 'nt' and not os.getenv('BB_NATIVE'):
        return Windows.browse_file(parent, title)

    chooser = Gtk.FileChooserDialog(title=title,
                                    transient_for=parent,
                                    action=Gtk.FileChooserAction.OPEN)
    chooser.add_buttons(CANCEL_BUTTON_LABEL, Gtk.ResponseType.CANCEL,
                        # TRANSLATORS: This is a label for the Open button in a file chooser dialog.
                        _("_Open"), Gtk.ResponseType.OK)
    chooser.set_default_response(Gtk.ResponseType.OK)
    chooser.set_current_folder(os.path.expanduser('~'))
    resp = chooser.run()
    path = chooser.get_filename()
    chooser.destroy()

    if Gtk.ResponseType.OK != resp:
        # user cancelled
        return None

    return path


def browse_files(parent, title):
    """Prompt user to select multiple files to delete"""

    if os.name == 'nt' and not os.getenv('BB_NATIVE'):
        return Windows.browse_files(parent, title)

    chooser = Gtk.FileChooserDialog(title=title,
                                    transient_for=parent,
                                    action=Gtk.FileChooserAction.OPEN)
    chooser.add_buttons(CANCEL_BUTTON_LABEL, Gtk.ResponseType.CANCEL,
                        DELETE_BUTTON_LABEL, Gtk.ResponseType.OK)
    chooser.set_default_response(Gtk.ResponseType.OK)
    chooser.set_select_multiple(True)
    chooser.set_current_folder(os.path.expanduser('~'))
    resp = chooser.run()
    paths = chooser.get_filenames()
    chooser.destroy()

    if Gtk.ResponseType.OK != resp:
        # user cancelled
        return None

    return paths


def delete_confirmation_dialog(parent, mention_preview, shred_settings=False):
    """Return boolean whether OK to delete files."""
    # TRANSLATORS: Title of the delete confirmation dialog.
    dialog = Gtk.Dialog(title=_("Delete confirmation"), transient_for=parent,
                        modal=True,
                        destroy_with_parent=True)
    dialog.set_default_size(300, -1)

    vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL,
                   homogeneous=False, spacing=10)

    if shred_settings:
        # TRANSLATORS: This message appears in a dialog window when the user
        # chooses to shred settings and quit.
        notice_text = _("This function deletes all BleachBit settings and then quits the "
                        "application. Use this to hide your use of BleachBit or to reset its "
                        "settings. The next time you start BleachBit, the settings will "
                        "initialize to default values.")
        notice = Gtk.Label(label=notice_text)
        notice.set_line_wrap(True)
        vbox.pack_start(notice, False, True, 0)

    if mention_preview:
        # TRANSLATORS: This message appears in a dialog window when the user
        # asks the user to confirm deleting files when a preview was run.
        question_text = _("Are you sure you want to permanently delete files "
                          "according to the selected operations?  The actual files that will be "
                          "deleted may have changed since you ran the preview.")
    else:
        # TRANSLATORS: This message appears in a dialog window when the user
        # asks the user to confirm deleting files when no preview was run.
        question_text = _("Are you sure you want to permanently delete "
                          "these files?")
    question = Gtk.Label(label=question_text)
    question.set_line_wrap(True)
    vbox.pack_start(question, False, True, 0)

    if options.get('expert_mode'):
        # TRANSLATORS: Check button shown in the confirm delete dialog.
        # If unchecked, then it will not confirm next time.
        cb_popup = Gtk.CheckButton(label=_("Confirm before delete"))
        cb_popup.set_active(options.get('delete_confirmation'))
        vbox.pack_start(cb_popup, False, True, 0)

    dialog.get_content_area().pack_start(vbox, False, True, 0)
    dialog.get_content_area().set_spacing(10)

    dialog.add_button(DELETE_BUTTON_LABEL, Gtk.ResponseType.ACCEPT)
    dialog.add_button(CANCEL_BUTTON_LABEL, Gtk.ResponseType.CANCEL)
    dialog.set_default_response(Gtk.ResponseType.CANCEL)

    dialog.show_all()
    ret = dialog.run()
    if options.get('expert_mode'):
        options.set('delete_confirmation', cb_popup.get_active())
    dialog.destroy()
    return ret == Gtk.ResponseType.ACCEPT


def warning_confirm_dialog(parent, option_name, warning_text, show_checkbox=True):
    """Show a warning dialog when enabling an option.

    Returns tuple (confirmed: bool, remember_choice: bool).
    """
    # TRANSLATORS: Title of a warning dialog. %(option)s is the name
    # of the option being enabled (e.g. "shred files").
    dialog = Gtk.Dialog(title=_('Enable %(option)s') % {'option': option_name},
                        transient_for=parent,
                        modal=True,
                        destroy_with_parent=True)
    content = dialog.get_content_area()
    content.set_border_width(12)
    content.set_spacing(10)

    warning_label = Gtk.Label(label=warning_text)
    warning_label.set_line_wrap(True)
    warning_label.set_max_width_chars(80)
    warning_label.set_xalign(0.0)
    content.pack_start(warning_label, False, False, 0)

    remember_cb = Gtk.CheckButton(
        # TRANSLATORS: Check button label. %(option)s is the name of
        # the option being enabled (e.g. "shred files").
        label=_('Remember my choice for %(option)s') % {'option': option_name})
    remember_cb.set_halign(Gtk.Align.START)
    if show_checkbox:
        content.pack_start(remember_cb, False, False, 0)

    cancel_button = dialog.add_button(
        CANCEL_BUTTON_LABEL, Gtk.ResponseType.CANCEL)
    # TRANSLATORS: Button label in a warning dialog to confirm enabling an option.
    # The underscore is the accelerator key.
    enable_button = dialog.add_button(_('_Enable anyway'), Gtk.ResponseType.OK)
    enable_button.get_style_context().add_class('destructive-action')
    dialog.set_default_response(Gtk.ResponseType.CANCEL)
    cancel_button.grab_focus()

    # Disable enable button for a moment to prevent accident by double-clicking.
    enable_button.set_sensitive(False)

    def enable_button_after_delay():
        enable_button.set_sensitive(True)
        return False

    GLib.timeout_add(500, enable_button_after_delay)

    dialog.show_all()
    response = dialog.run()
    remember_choice = response == Gtk.ResponseType.OK and show_checkbox and remember_cb.get_active()
    dialog.destroy()

    return response == Gtk.ResponseType.OK, remember_choice


def message_dialog(parent, msg, mtype=Gtk.MessageType.ERROR, buttons=Gtk.ButtonsType.OK, title=None):
    """Convenience wrapper for Gtk.MessageDialog"""

    dialog = Gtk.MessageDialog(transient_for=parent,
                               modal=True,
                               destroy_with_parent=True,
                               message_type=mtype,
                               buttons=buttons,
                               text=msg)
    if title:
        dialog.set_title(title)
    resp = dialog.run()
    dialog.destroy()

    return resp


def open_url(url, parent_window=None, prompt=True):
    """Open an HTTP URL.  Try to run as non-root."""
    # drop privileges so the web browser is running as a normal process
    if os.name == 'posix' and os.getuid() == 0:
        # TRANSLATORS: This is an error message shown to root users.
        # %s expands to a web URL.
        msg = _("Because you are running as root, please manually open "
                "this link in a web browser:\n%s") % url
        message_dialog(None, msg, Gtk.MessageType.INFO)
        return
    if prompt:
        # find hostname
        import re
        ret = re.search(r'^http(s)?://([a-z.]+)', url)
        if not ret:
            host = url
        else:
            host = ret.group(2)
        # TRANSLATORS: The question appears in a confirmation dialog.
        # %s expands to www.bleachbit.org or similar
        msg = _("Open web browser to %s?") % host
        resp = message_dialog(parent_window,
                              msg,
                              Gtk.MessageType.QUESTION,
                              Gtk.ButtonsType.OK_CANCEL,
                              # TRANSLATORS: Title of a confirmation dialog.
                              _('Confirm'))
        if Gtk.ResponseType.OK != resp:
            return
    # open web browser
    if os.name == 'nt':
        # in Gtk.show_uri() avoid 'glib.GError: No application is registered as
        # handling this file'
        import webbrowser
        webbrowser.open(url)
    elif (Gtk.get_major_version(), Gtk.get_minor_version()) < (3, 22):
        # Ubuntu 16.04 LTS ships with GTK 3.18
        Gtk.show_uri(None, url, Gdk.CURRENT_TIME)
    else:
        Gtk.show_uri_on_window(parent_window, url, Gdk.CURRENT_TIME)



# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Store and retrieve user preferences
"""

# standard library imports
import atexit
import errno
import logging
import os
import re
import threading

# third-party imports
if 'nt' == os.name:
    from win32file import GetLongPathName

# local application imports
import bleachbit
from bleachbit import General
from bleachbit.Language import get_text as _

logger = logging.getLogger(__name__)

FLUSH_DELAY_SECS = 15.0  # decimal seconds

OPTION_DEFAULTS = {
    'auto_hide': {'value': True},
    'auto_detect_lang': {'value': True},
    'check_beta': {'value': False},
    'check_online_updates': {'value': True},
    'dark_mode': {'value': True},
    'debug': {'value': False},
    'delete_confirmation': {'value': True},
    'expert_mode': {'value': False},
    'exit_done': {'value': False},
    'first_start': {'value': False},
    'kde_shred_menu_option': {'value': False},
    'load_cleaners': {'value': True},
    'remember_geometry': {'value': True},
    'shred': {'value': False},
    'units_iec': {'value': False},
    'window_maximized': {'value': False},
    'window_fullscreen': {'value': False},
    'font_check_completed': {'value': False, 'platforms': ('nt',)},
    'update_winapp2': {'value': False, 'platforms': ('nt',)},
    'use_fontconfig_backend': {'value': False, 'platforms': ('nt',)},
    'win10_theme': {'value': False, 'platforms': ('nt',)},
}


def _platform_allows(meta):
    platforms = meta.get('platforms')
    return not platforms or os.name in platforms


def _get_default_value(option):
    meta = OPTION_DEFAULTS.get(option)
    if not meta or not _platform_allows(meta):
        return None
    return meta['value']


boolean_keys = [
    key for key, meta in OPTION_DEFAULTS.items()
    if _platform_allows(meta)
]
int_keys = ['window_x', 'window_y', 'window_width', 'window_height', ]


def _option_index(option_name):
    """Extract the numeric index from an option name.

    Handles both plain integer keys (e.g. '0', '1') used by get_list
    and compound keys (e.g. '1_type', '2_path') used by get_paths.
    """
    return int(option_name.split('_')[0])


def path_to_option(pathname):
    """Change a pathname to a .ini option name (a key)"""
    # On Windows change to lowercase and use backwards slashes.
    pathname = os.path.normcase(pathname)
    # On Windows expand DOS-8.3-style pathnames.
    if 'nt' == os.name and os.path.exists(pathname):
        pathname = GetLongPathName(pathname)
    if len(pathname) > 1 and ':' == pathname[1]:
        # ConfigParser treats colons in a special way
        pathname = pathname[0] + pathname[2:]
    return pathname


def init_configuration(*, log=True):
    """Initialize an empty configuration, if necessary"""
    if not os.path.exists(bleachbit.options_dir):
        General.makedirs(bleachbit.options_dir)
    if os.path.lexists(bleachbit.options_file):
        if log:
            logger.debug('Deleting configuration: %s', bleachbit.options_file)
        os.remove(bleachbit.options_file)
    with open(bleachbit.options_file, 'w', encoding='utf-8-sig', errors='surrogateescape') as f_ini:
        f_ini.write('[bleachbit]\n')
        if os.name == 'nt' and bleachbit.portable_mode:
            f_ini.write('[Portable]\n')
    for section in options.config.sections():
        options.config.remove_section(section)
    options.restore()


class Options:

    """Store and retrieve user preferences"""

    def __init__(self):
        self.purged = False
        self.config = bleachbit.RawConfigParser()
        self.config.optionxform = str  # make keys case sensitive for hashpath purging
        self.config.BOOLEAN_STATES['t'] = True
        self.config.BOOLEAN_STATES['f'] = False
        self.overrides = {}
        self.old_version = None  # Store previous version in memory
        self._dirty = False
        self._closed = False
        self._flush_generation = 0
        self._flush_lock = threading.RLock()
        self._flush_timer = None
        self._atexit_close = self.close
        atexit.register(self._atexit_close)
        self.restore()

        old_option = 'system.free_disk_space'
        try:
            if self.config.has_section("tree") and self.config.has_option('tree', old_option):
                logger.debug(
                    "Migrating legacy option '%s' to 'system.empty_space'", old_option)
                self.config.set('tree', 'system.empty_space', 'true')
                self.config.remove_option('tree', old_option)
                self.__schedule_flush()
        except Exception:
            logger.exception("Error migrating legacy option '%s'", old_option)

    def __cancel_flush_timer(self):
        """Invalidate and stop any pending delayed flush timer."""
        self._flush_generation += 1
        timer = self._flush_timer
        self._flush_timer = None
        if timer is not None:
            timer.cancel()

    def __flush_after_delay(self, generation):
        """Timer callback that flushes if the timer is still current."""
        with self._flush_lock:
            if generation != self._flush_generation:
                return
            self._flush_timer = None
            self.__flush_locked()

    def __schedule_flush(self):
        """Mark configuration dirty and schedule a delayed flush."""
        with self._flush_lock:
            self._dirty = True
            self._flush_generation += 1
            generation = self._flush_generation
            if self._flush_timer is not None:
                self._flush_timer.cancel()
            self._flush_timer = threading.Timer(
                FLUSH_DELAY_SECS, self.__flush_after_delay, args=(generation,))
            self._flush_timer.daemon = True
            self._flush_timer.start()

    def __flush(self, force=False):
        """Acquire the lock and flush to disk."""
        with self._flush_lock:
            self.__flush_locked(force=force)

    def __flush_locked(self, force=False):
        """Write configuration to disk while holding the flush lock."""
        if not force and not self._dirty:
            return
        if not self.purged:
            self.__purge()

        try:
            if not os.path.exists(bleachbit.options_dir):
                General.makedirs(bleachbit.options_dir)
            mkfile = not os.path.exists(bleachbit.options_file)
            with open(bleachbit.options_file, 'w', encoding='utf-8-sig', errors='surrogateescape') as _file:
                self.config.write(_file)
            if mkfile and General.sudo_mode():
                General.chownself(bleachbit.options_file)
        except (OSError, IOError, PermissionError) as e:
            if e.errno == errno.ENOSPC:
                logger.error(
                    _("Disk was full when writing configuration to file: %s"), bleachbit.options_file)
            elif e.errno == errno.EACCES:
                logger.error(
                    _("Permission denied when writing configuration to file: %s"), bleachbit.options_file)
            else:
                raise
        else:
            self._dirty = False

    def __purge(self):
        """Clear out obsolete data"""
        self.purged = True
        if not self.config.has_section('hashpath'):
            return
        for option in self.config.options('hashpath'):
            pathname = option
            if 'nt' == os.name and re.search(r'^[a-z]\\', option):
                # restore colon lost because ConfigParser treats colon special
                # in keys
                pathname = pathname[0] + ':' + pathname[1:]
            exists = False
            try:
                exists = os.path.lexists(pathname)
            except:
                # this deals with corrupt keys
                # https://www.bleachbit.org/forum/bleachbit-wont-launch-error-startup
                logger.error(
                    _("Error checking whether path exists: %s"), pathname)
            if not exists:
                # the file does not on exist, so forget it
                self.config.remove_option('hashpath', option)

    def __auto_preserve_languages(self):
        """Automatically preserve the active language"""
        active_lang = bleachbit.Language.get_active_language_code()
        for lang_id in set([active_lang.split('_')[0], 'en']):
            logger.info(_("Automatically preserving language %s."), lang_id)
            self.set_language(lang_id, True)

    def has_option(self, option, section='bleachbit'):
        """Check if option is set"""
        return self.config.has_option(section, option)

    def has_override(self, option, section='bleachbit'):
        """Check if option is overridden"""
        return (section, option) in self.overrides

    def get(self, option, section='bleachbit'):
        """Retrieve a general option"""
        if not 'nt' == os.name and 'update_winapp2' == option:
            return False
        if section == 'bleachbit' and option == 'debug':
            from bleachbit.Log import is_debugging_enabled_via_cli
            if is_debugging_enabled_via_cli():
                # command line overrides stored configuration
                return True
        override_key = (section, option)
        if override_key in self.overrides:
            return self.overrides[override_key]
        if section == 'hashpath' and len(option) > 1 and option[1] == ':':
            option = option[0] + option[2:]
        if self.config.has_option(section, option):
            if option in boolean_keys:
                return self.config.getboolean(section, option)
            if option in int_keys:
                return self.config.getint(section, option)
            return self.config.get(section, option)

        if section == 'bleachbit':
            default = _get_default_value(option)
            if default is not None:
                return default

        return None

    def get_hashpath(self, pathname):
        """Recall the hash for a file"""
        return self.get(path_to_option(pathname), 'hashpath')

    def get_language(self, langid):
        """Retrieve value for whether to preserve the language"""
        if not self.config.has_section('preserve_languages'):
            self.__auto_preserve_languages()
        if not self.config.has_option('preserve_languages', langid):
            return False
        return self.config.getboolean('preserve_languages', langid)

    def get_languages(self):
        """Return a list of all selected languages"""
        if not self.config.has_section('preserve_languages'):
            self.__auto_preserve_languages()
        return self.config.options('preserve_languages')

    def get_list(self, option):
        """Return an option which is a list data type"""
        section = f"list/{option}"
        if not self.config.has_section(section):
            return None
        values = [
            self.config.get(section, opt)
            for opt in sorted(self.config.options(section), key=_option_index)
        ]
        return values

    def get_paths(self, section):
        """Abstracts get_whitelist_paths and get_custom_paths"""
        if not self.config.has_section(section):
            return []
        myoptions = []
        for opt in sorted(self.config.options(section), key=_option_index):
            pos = opt.find('_')
            if -1 == pos:
                continue
            myoptions.append(opt[0:pos])
        values = []
        for opt in sorted(set(myoptions), key=_option_index):
            p_type = self.config.get(section, opt + '_type')
            p_path = self.config.get(section, opt + '_path')
            values.append((p_type, p_path))
        return values

    def get_whitelist_paths(self):
        """Return the keep list (formerly whitelist) of paths

        Returns a list of tuples (type, path) where type is 'file' or 'folder'
        """
        return self.get_paths("whitelist/paths")

    def get_custom_paths(self):
        """Return list of custom paths

        Returns a list of tuples (type, path) where type is 'file' or 'folder'
        """
        return self.get_paths("custom/paths")

    def get_warning_preference(self, key):
        """Return whether a specific warning is already confirmed."""
        section = "warnings"
        if not self.config.has_section(section):
            return False
        if not self.config.has_option(section, key):
            return False
        try:
            return self.config.getboolean(section, key)
        except ValueError:
            logger.exception("Error reading warning preference for %s", key)
            return False

    def remember_warning_preference(self, key):
        """Persist that a specific warning was confirmed once."""
        section = "warnings"
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, key, 'True')
        self.__schedule_flush()

    def clear_warning_preferences(self):
        """Clear all saved warning confirmations."""
        section = "warnings"
        if self.config.has_section(section):
            self.config.remove_section(section)
            self.__schedule_flush()

    def get_tree(self, parent, child):
        """Retrieve an option for the tree view.

        Args:
            parent: The cleaner ID (e.g., 'system', 'firefox')
            child: The option ID within the cleaner (e.g., 'cache'), or None for the cleaner itself

        Returns:
            bool: True if the option is enabled, False otherwise
        """
        option = parent
        if child is not None:
            option += "." + child

        if not self.config.has_option('tree', option):
            return False
        try:
            return self.config.getboolean('tree', option)
        except:
            # in case of corrupt configuration (Launchpad #799130)
            logger.exception('Error in get_tree()')
            return False

    def is_corrupt(self):
        """Perform a self-check for corruption of the configuration"""
        # no boolean key must raise an exception
        for boolean_key in boolean_keys:
            try:
                if self.config.has_option('bleachbit', boolean_key):
                    self.config.getboolean('bleachbit', boolean_key)
            except ValueError:
                return True
        # no int key must raise an exception
        for int_key in int_keys:
            try:
                if self.config.has_option('bleachbit', int_key):
                    self.config.getint('bleachbit', int_key)
            except ValueError:
                return True
        return False

    def reset_overrides(self):
        """Clear all overrides"""
        self.overrides.clear()

    def restore(self):
        """Restore saved options from disk

        Abandons any changes not written to disk.
        """
        with self._flush_lock:
            self.__cancel_flush_timer()
            self._dirty = False
            # Reading configuration merges with existing data,
            # so clear it first.
            for section in self.config.sections():
                self.config.remove_section(section)
        try:
            with open(bleachbit.options_file, 'r', encoding='utf-8-sig', errors='surrogateescape') as _file:
                self.config.read_file(_file, bleachbit.options_file)
        except FileNotFoundError:
            logger.debug("Configuration file does not exist yet: %s",
                         bleachbit.options_file)
        except Exception:
            logger.exception("Error reading application's configuration")
        if not self.config.has_section("bleachbit"):
            self.config.add_section("bleachbit")
        if not self.config.has_section("hashpath"):
            self.config.add_section("hashpath")
        if not self.config.has_section("list/shred_drives"):
            from bleachbit.FileUtilities import guess_overwrite_paths
            try:
                self.set_list('shred_drives', guess_overwrite_paths())
            except:
                logger.exception(
                    _("Error when setting the default drives to shred."))
        # BleachBit upgrade or first start ever
        if not self.config.has_option('bleachbit', 'version') or \
                self.get('version') != bleachbit.APP_VERSION:
            if self.config.has_option('bleachbit', 'version'):
                self.old_version = self.get('version')
            self.set('first_start', True)

        # set version
        self.set("version", bleachbit.APP_VERSION)

    def set(self, key, value, section='bleachbit'):
        """Set a general option"""
        assert isinstance(key, str), f"key must be a string: {key}"
        assert isinstance(section, str), f"section must be a string: {section}"
        override_key = (section, key)
        if override_key not in self.overrides:
            value = str(value)
            if self.config.has_option(section, key) and self.config.get(section, key) == value:
                return
            self.config.set(section, key, value)
            self.__schedule_flush()

    def commit(self):
        """Cancel times and write changes"""
        with self._flush_lock:
            self.__cancel_flush_timer()
            self.__flush(force=True)

    def close(self):
        """Cancel times and write changes"""
        with self._flush_lock:
            if self._closed:
                return
            self._closed = True
            atexit.unregister(self._atexit_close)
            self.__cancel_flush_timer()
            self.__flush(force=False)

    def set_hashpath(self, pathname, hashvalue):
        """Remember the hash of a path"""
        self.set(path_to_option(pathname), hashvalue, 'hashpath')

    def set_list(self, key, values):
        """Set a value which is a list data type"""
        assert isinstance(key, str), f"key must be a string: {key}"
        section = f"list/{key}"
        # Remove existing section first to clear old values
        # before writing new ones.
        if self.config.has_section(section):
            self.config.remove_section(section)
        self.config.add_section(section)
        for counter, value in enumerate(values):
            self.config.set(section, str(counter), value)
        self.__schedule_flush()

    def set_whitelist_paths(self, values):
        """Save the keep list (formerly whitelist)"""
        section = "whitelist/paths"
        # Remove existing section first to clear old values
        # before writing new ones.
        if self.config.has_section(section):
            self.config.remove_section(section)
        self.config.add_section(section)
        for counter, value in enumerate(values):
            self.config.set(section, str(counter) + '_type', value[0])
            self.config.set(section, str(counter) + '_path', value[1])
        self.__schedule_flush()

    def set_custom_paths(self, values):
        """Save the custom paths

        @param values: list of tuples containing (path_type, path)
            where path_type is either 'file' or 'folder'
        """
        section = "custom/paths"
        # Remove existing section first to clear old values
        # before writing new ones.
        if self.config.has_section(section):
            self.config.remove_section(section)
        self.config.add_section(section)
        for counter, value in enumerate(values):
            path_type, path = value
            assert path_type in ('file', 'folder')
            self.config.set(section, str(counter) + '_type', path_type)
            self.config.set(section, str(counter) + '_path', path)
        self.__schedule_flush()

    def set_language(self, langid, value):
        """Set the value for a locale (whether to preserve it)"""
        if not self.config.has_section('preserve_languages'):
            self.config.add_section('preserve_languages')
        if self.config.has_option('preserve_languages', langid) and not value:
            self.config.remove_option('preserve_languages', langid)
        else:
            self.config.set('preserve_languages', langid, str(value))
        self.__schedule_flush()

    def set_tree(self, parent, child, value):
        """Set an option for the tree view.  The child may be None."""
        if not self.config.has_section("tree"):
            self.config.add_section("tree")
        option = parent
        if child is not None:
            option = option + "." + child
        if self.config.has_option('tree', option) and not value:
            self.config.remove_option('tree', option)
        else:
            self.config.set('tree', option, str(value))
        self.__schedule_flush()

    def toggle(self, key):
        """Toggle a boolean key"""
        self.set(key, not self.get(key))

    def set_override(self, key, value, section='bleachbit'):
        """Set a CLI override that will never be written to disk"""
        override_key = (section, key)
        self.overrides[override_key] = value


options = Options()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
GUI for making chaff
"""

import functools
import logging
import os
import shutil
import tempfile
import threading


from bleachbit.Chaff import generate_emails, generate_2600
from bleachbit.Constant import ABORT_BUTTON_LABEL
from bleachbit.GtkShim import Gtk, GLib
from bleachbit.Language import get_text as _


logger = logging.getLogger(__name__)


STOP_MODE_FILE_COUNT = 0
STOP_MODE_TOTAL_SIZE = 1
STOP_MODE_FREE_SPACE = 2

MAX_FILE_COUNT = 999999
MAX_MB_COUNT = 999999
MAX_FREE_SPACE_PCT = 99

STOP_MODE_LABELS = {
    # TRANSLATORS: Option in combo box to choose to stop after a number
    # of files have been generated.
    STOP_MODE_FILE_COUNT: _("Number of files"),
    # TRANSLATORS: Option in a combo box for choosing the stop condition
    # when generating chaff (dummy) files. This option means "stop after
    # the total size of generated files reaches a threshold in megabytes."
    # "MB" is the megabyte unit abbreviation. Localize it if your language
    # has an official abbreviation (e.g., "Mo" in French); otherwise keep
    # "MB".
    STOP_MODE_TOTAL_SIZE: _("Size (MB)"),
    # TRANSLATORS: Option in combo box to choose to stop when free space
    # reaches a certain percentage.
    # The percentage symbol (%) is a literal character, not a format
    # placeholder.
    STOP_MODE_FREE_SPACE: _("Free space (%)"),
}


# TRANSLATORS: Used for (1) label for folder chooser button and
# (2) error message shown in the infobar when the folder has not
# been set. 'Select' is a verb.
SELECT_DEST_FOLDER_MSG = _("Select destination folder")


def _make_should_stop(stop_mode, stop_value, output_folder, abort_event):
    """Create a should_stop callback based on the stop mode.

    The returned callable accepts generated_file_names (list of paths)
    and cumulative_size (total bytes written so far).

    Returns (should_stop, file_count).
    """
    if stop_mode == STOP_MODE_FILE_COUNT:
        def should_stop(generated_file_names, cumulative_size=0):  # pylint: disable=unused-argument
            return abort_event.is_set()

        return should_stop, stop_value

    if stop_mode == STOP_MODE_TOTAL_SIZE:
        target_bytes = stop_value * 1024 * 1024  # MB to bytes

        def should_stop(generated_file_names, cumulative_size=0):  # pylint: disable=unused-argument
            if abort_event.is_set():
                return True
            return cumulative_size >= target_bytes

        return should_stop, MAX_FILE_COUNT

    if stop_mode == STOP_MODE_FREE_SPACE:
        target_free_pct = stop_value

        def should_stop(generated_file_names, cumulative_size=0):  # pylint: disable=unused-argument
            if abort_event.is_set():
                return True
            try:
                usage = shutil.disk_usage(output_folder)
            except FileNotFoundError:
                return False
            free_pct = 100.0 * usage.free / usage.total
            return free_pct <= target_free_pct

        return should_stop, MAX_FILE_COUNT

    raise ValueError(f'Invalid stop_mode {stop_mode}')


def _make_progress_cb(stop_mode, stop_value, output_folder, on_progress):
    """Create a progress callback appropriate for the stop mode.

    For file count mode, the fraction from Chaff is already correct.
    For size/free-space modes, we use cumulative_size passed from Chaff
    or compute from disk usage.
    """
    if stop_mode == STOP_MODE_FILE_COUNT:
        def progress_cb(fraction, generated_file_names=None, cumulative_size=0):  # pylint: disable=unused-argument
            on_progress(fraction)

        return progress_cb

    if stop_mode == STOP_MODE_TOTAL_SIZE:
        target_bytes = stop_value * 1024 * 1024

        def progress_cb(fraction, generated_file_names=None, cumulative_size=0):  # pylint: disable=unused-argument
            if cumulative_size > 0:
                on_progress(min(1.0, cumulative_size / target_bytes))
            else:
                on_progress(fraction)

        return progress_cb

    if stop_mode == STOP_MODE_FREE_SPACE:
        target_free_pct = stop_value
        initial_free_pct = [None]  # Use list to allow mutation in nested function

        def progress_cb(_fraction, generated_file_names=None, cumulative_size=0):  # pylint: disable=unused-argument
            try:
                usage = shutil.disk_usage(output_folder)
            except FileNotFoundError:
                on_progress(0.0)
                return
            current_free_pct = 100.0 * usage.free / usage.total
            if initial_free_pct[0] is None:
                initial_free_pct[0] = current_free_pct
            if initial_free_pct[0] > target_free_pct:
                frac = (initial_free_pct[0] - current_free_pct) / \
                    (initial_free_pct[0] - target_free_pct)
                frac = min(1.0, max(0.0, frac))
            else:
                frac = 1.0
            on_progress(frac)

        return progress_cb

    raise ValueError(f'Invalid stop_mode {stop_mode}')


def make_files_thread(stop_mode, stop_value, inspiration, output_folder,
                      delete_when_finished, on_progress, abort_event):
    """Make files in a separate thread"""
    should_stop, file_count = _make_should_stop(
        stop_mode, stop_value, output_folder, abort_event)
    progress_cb = _make_progress_cb(
        stop_mode, stop_value, output_folder, on_progress)

    try:
        if inspiration == 0:
            generated_file_names = generate_2600(
                file_count, output_folder, on_progress=progress_cb,
                should_stop=should_stop)
        elif inspiration == 1:
            generated_file_names = generate_emails(
                file_count, output_folder, on_progress=progress_cb,
                should_stop=should_stop)
        else:
            raise ValueError(f'Invalid inspiration {inspiration}')
    except Exception as exc:
        logger.exception('Error generating chaff')
        # TRANSLATORS: Error message shown when chaff file generation fails.
        # The placeholder is for the technical error details.
        error_msg = _("Error generating chaff: {error}").format(error=str(exc))
        on_progress(1.0, is_done=True, error=error_msg)
        return
    try:
        if delete_when_finished and not abort_event.is_set():
            # TRANSLATORS: Progress message shown while deleting chaff files.
            # 'Deleting files' is a present participle.
            # To indicate an ongoing operation, include the ellipsis as literal
            # Unicode (…) or as Unicode escape (\u2026).
            on_progress(0, msg=_('Deleting files\u2026'))
            count = len(generated_file_names)
            for i, fn in enumerate(generated_file_names):
                if abort_event.is_set():
                    break
                os.unlink(fn)
                on_progress(1.0 * (i + 1) / count)
    except Exception as exc:
        logger.exception('Error deleting chaff files')
        # TRANSLATORS: Error message shown when deleting chaff files fails.
        # The placeholder is for the technical error details.
        error_msg = _("Error deleting chaff files: {error}").format(error=str(exc))
        on_progress(1.0, is_done=True, error=error_msg)
        return
    on_progress(1.0, is_done=True)


class ChaffDialog(Gtk.Dialog):

    """Present the dialog to make chaff"""

    _infobar_timeout_id = None
    _download_success = None
    _abort_event = None
    thread = None

    def __init__(self, parent):
        self._make_dialog(parent)

    def _make_dialog(self, parent):
        """Make the main dialog"""
        # TRANSLATORS: Title for dialog window.
        # Digital chaff is like physical chaff that airplanes use to protect themselves
        # from radar-guided missiles. For more explanation, see the online documentation.
        Gtk.Dialog.__init__(self, title=_("Make chaff"), transient_for=parent)
        Gtk.Dialog.set_modal(self, True)
        self.set_border_width(10)
        self.set_default_size(400, -1)
        self.connect('delete-event', self._on_delete_event)
        box = self.get_content_area()
        box.set_spacing(10)

        # Add InfoBar for non-blocking messages
        self.infobar = Gtk.InfoBar()
        self.infobar.set_show_close_button(True)
        self.infobar.connect('response', self._on_infobar_response)
        self.infobar_label = Gtk.Label()
        self.infobar_label.set_line_wrap(True)
        self.infobar.get_content_area().add(self.infobar_label)
        box.pack_start(self.infobar, False, False, 0)
        self._infobar_timeout_id = None

        # TRANSLATORS: Label at the top of the chaff dialog
        dialog_label = _("Make randomly-generated messages "
                         "inspired by documents.")
        label = Gtk.Label(label=dialog_label)
        label.set_line_wrap(True)
        label.set_xalign(0)
        box.pack_start(label, False, False, 0)

        grid = Gtk.Grid()
        grid.set_column_spacing(12)
        grid.set_row_spacing(8)
        box.pack_start(grid, False, False, 0)

        # TRANSLATORS: Label for the inspiration combo box.
        # 'Inspiration' is a choice of documents from which random text will be generated.
        inspiration_label = Gtk.Label(label=_("Inspiration"))
        inspiration_label.set_xalign(0)
        grid.attach(inspiration_label, 0, 0, 1, 1)
        self.inspiration_combo = Gtk.ComboBoxText()
        self.inspiration_combo.set_hexpand(True)
        self.inspiration_combo_options = (
            _('2600 Magazine'), _("Hillary Clinton's emails"))
        for combo_option in self.inspiration_combo_options:
            self.inspiration_combo.append_text(combo_option)
        self.inspiration_combo.set_active(0)  # Set default
        grid.attach(self.inspiration_combo, 1, 0, 1, 1)

        # TRANSLATORS: Label for the combo box that selects when to stop
        # generating chaff files.
        stop_after_label = Gtk.Label(label=_("Stop after"))
        stop_after_label.set_xalign(0)
        grid.attach(stop_after_label, 0, 1, 1, 1)
        self.stop_mode_combo = Gtk.ComboBoxText()
        self.stop_mode_combo.set_hexpand(True)
        for mode in (STOP_MODE_FILE_COUNT, STOP_MODE_TOTAL_SIZE, STOP_MODE_FREE_SPACE):
            self.stop_mode_combo.append_text(STOP_MODE_LABELS[mode])
        self.stop_mode_combo.set_active(STOP_MODE_FILE_COUNT)
        self.stop_mode_combo.connect('changed', self._on_stop_mode_changed)
        grid.attach(self.stop_mode_combo, 1, 1, 1, 1)

        # TRANSLATORS: Label for the spin button that selects the stop value.
        self.stop_value_label = Gtk.Label(
            label=STOP_MODE_LABELS[self.stop_mode_combo.get_active()])
        self.stop_value_label.set_xalign(0)
        grid.attach(self.stop_value_label, 0, 2, 1, 1)
        self._stop_value_adjustments = {
            STOP_MODE_FILE_COUNT: Gtk.Adjustment(
                value=100, lower=1, upper=MAX_FILE_COUNT,
                step_increment=1, page_increment=1000, page_size=0),
            STOP_MODE_TOTAL_SIZE: Gtk.Adjustment(
                value=100, lower=1, upper=MAX_MB_COUNT,
                step_increment=100, page_increment=1000, page_size=0),
            STOP_MODE_FREE_SPACE: Gtk.Adjustment(
                value=1, lower=1, upper=MAX_FREE_SPACE_PCT,
                step_increment=1, page_increment=10, page_size=0),
        }
        self.stop_value_spin = Gtk.SpinButton(
            adjustment=self._stop_value_adjustments[STOP_MODE_FILE_COUNT])
        self.stop_value_spin.set_hexpand(True)
        grid.attach(self.stop_value_spin, 1, 2, 1, 1)

        folder_label = Gtk.Label(label=SELECT_DEST_FOLDER_MSG)
        folder_label.set_xalign(0)
        grid.attach(folder_label, 0, 3, 1, 1)
        # The file chooser button displays a stock GTK icon. When some parts of GTK are not
        # set up correctly on Windows, then the application may crash here with the error
        # message "No GSettings schemas".
        # https://github.com/bleachbit/bleachbit/issues/1780
        self.choose_folder_button = Gtk.FileChooserButton()
        self.choose_folder_button.set_action(
            Gtk.FileChooserAction.SELECT_FOLDER)
        self.choose_folder_button.set_filename(tempfile.gettempdir())
        self.choose_folder_button.set_hexpand(True)
        grid.attach(self.choose_folder_button, 1, 3, 1, 1)

        # TRANSLATORS: Label for the combo box that selects what to do
        # when chaff generation is finished.
        finished_label = Gtk.Label(label=_("When finished"))
        finished_label.set_xalign(0)
        grid.attach(finished_label, 0, 4, 1, 1)
        self.when_finished_combo = Gtk.ComboBoxText()
        self.when_finished_combo.set_hexpand(True)
        self.combo_options = (
            # TRANSLATORS: Option in combo box to delete generated chaff
            # files without shredding them.
            _('Delete without shredding'),
            # TRANSLATORS: Option in combo box to keep generated chaff files.
            _('Do not delete'))
        for combo_option in self.combo_options:
            self.when_finished_combo.append_text(combo_option)
        self.when_finished_combo.set_active(0)  # Set default
        grid.attach(self.when_finished_combo, 1, 4, 1, 1)

        # Loading indicator for download (hidden by default)
        self._download_spinner_box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        self._download_spinner_box.set_halign(Gtk.Align.CENTER)
        self._download_spinner = Gtk.Spinner()
        self._download_spinner.set_size_request(24, 24)
        self._download_spinner_box.pack_start(
            self._download_spinner, False, False, 0)
        # TRANSLATORS: This is a label in a dialog shown when the user
        # clicks the button to download models for chaff generation.
        # 'Downloading' is a present participle.
        # 'Inspiration content' refers to source documents from which
        # random text will be generated.
        # To indicate an ongoing operation, include the ellipsis as literal
        # Unicode (…) or as Unicode escape (\u2026).
        download_label_str = _("Downloading inspiration content\u2026")
        self._download_label = Gtk.Label(label=download_label_str)
        self._download_spinner_box.pack_start(
            self._download_label, False, False, 0)
        box.pack_start(self._download_spinner_box, False, False, 0)
        self._download_spinner_box.set_no_show_all(True)
        self._download_spinner_box.hide()

        self.progressbar = Gtk.ProgressBar()
        box.pack_start(self.progressbar, False, False, 0)
        self.progressbar.hide()

        # TRANSLATORS: Button label in a dialog window to start making
        # chaff files.
        self.make_button = Gtk.Button(label=_("Make files"))
        self.make_button.get_style_context().add_class('suggested-action')
        self.make_button.connect('clicked', self.on_make_files)
        box.pack_start(self.make_button, False, False, 0)

        self.abort_button = Gtk.Button(label=ABORT_BUTTON_LABEL)
        self.abort_button.get_style_context().add_class('destructive-action')
        self.abort_button.connect('clicked', self._on_abort)
        box.pack_start(self.abort_button, False, False, 0)
        self.abort_button.set_sensitive(False)

        self._abort_event = None

    def _on_infobar_response(self, _infobar, _response_id):
        """Handle InfoBar close button click"""
        if self._infobar_timeout_id:
            GLib.source_remove(self._infobar_timeout_id)
            self._infobar_timeout_id = None
        self.infobar.hide()

    def _on_stop_mode_changed(self, combo):
        """Update the value spin button when the stop mode changes"""
        mode = combo.get_active()
        self.stop_value_label.set_text(STOP_MODE_LABELS[mode])
        self.stop_value_spin.set_adjustment(
            self._stop_value_adjustments[mode])

    def _on_abort(self, _widget):
        """Callback for abort button"""
        if self._abort_event:
            self._abort_event.set()

    def _on_delete_event(self, _widget, _event):
        """Handle dialog close (e.g., X button) by aborting the thread."""
        if self._abort_event:
            self._abort_event.set()
        return False  # Allow the dialog to close

    def _hide_infobar(self):
        """Hide the InfoBar (used for auto-dismiss timeout)"""
        self._infobar_timeout_id = None
        self.infobar.hide()
        return False  # Remove from GLib timeout

    def show_infobar(self, message, message_type=Gtk.MessageType.ERROR):
        """Show a non-blocking InfoBar message that auto-dismisses"""
        if self._infobar_timeout_id:
            GLib.source_remove(self._infobar_timeout_id)
            self._infobar_timeout_id = None
        self.infobar_label.set_text(message)
        self.infobar.set_message_type(message_type)
        self.infobar.show_all()
        self._infobar_timeout_id = GLib.timeout_add_seconds(
            15, self._hide_infobar)

    def download_models_gui(self, on_complete):
        """Download models in a background thread.

        Shows a spinner in the main dialog during download.
        Calls on_complete(success) when done, where success is boolean.
        """

        self._download_success = None

        def on_download_error(msg, msg2):
            # Use idle_add to show error from main thread
            message = f'{msg}: {msg2}'
            GLib.idle_add(
                functools.partial(self.show_infobar, message, Gtk.MessageType.ERROR))

        def download_models_thread(on_error):
            """Download models in a background thread."""
            import bleachbit.Chaff
            return bleachbit.Chaff.download_models(on_error=on_error)

        def _finish_download(success):
            """Called on main thread when download completes."""
            self._download_spinner.stop()
            self._download_spinner_box.hide()
            self._download_spinner_box.set_no_show_all(True)
            self.make_button.set_sensitive(True)
            on_complete(success)
            return False

        def on_thread_complete(success):
            """Callback when download thread completes."""
            GLib.idle_add(_finish_download, success)

        # Show loading state
        self._download_spinner.start()
        self._download_spinner_box.set_no_show_all(False)
        self._download_spinner_box.show_all()
        self.make_button.set_sensitive(False)

        # Start download in background thread
        def _worker():
            success = download_models_thread(on_download_error)
            on_thread_complete(success)

        thread = threading.Thread(target=_worker)
        thread.start()

    def on_make_files(self, _widget):
        """Callback for make files button"""
        self.infobar.hide()
        stop_mode = self.stop_mode_combo.get_active()
        stop_value = self.stop_value_spin.get_value_as_int()
        output_dir = self.choose_folder_button.get_filename()
        delete_when_finished = self.when_finished_combo.get_active() == 0
        inspiration = self.inspiration_combo.get_active()
        if not output_dir:
            self.show_infobar(SELECT_DEST_FOLDER_MSG,
                              Gtk.MessageType.ERROR)
            return

        from bleachbit.Chaff import have_models
        if not have_models():
            # Download models first, then proceed to file generation
            def on_download_complete(success):
                if success:
                    self._start_file_generation(
                        stop_mode, stop_value, inspiration, output_dir, delete_when_finished)
                else:
                    # TRANSLATORS: Error message shown when downloading
                    # chaff models failed.
                    self.show_infobar(_("Download failed"),
                                      Gtk.MessageType.ERROR)
            self.download_models_gui(on_download_complete)
        else:
            self._start_file_generation(
                stop_mode, stop_value, inspiration, output_dir, delete_when_finished)

    def _start_file_generation(self, stop_mode, stop_value, inspiration,
                               output_dir, delete_when_finished):
        """Start generating files after download is complete."""
        self._abort_event = threading.Event()

        def _on_progress(fraction, msg, is_done, error=None):
            """Update progress bar from GLib main loop"""
            if msg:
                self.progressbar.set_text(msg)
            self.progressbar.set_fraction(fraction)
            if is_done:
                self.progressbar.hide()
                self.abort_button.set_sensitive(False)
                self.make_button.set_sensitive(True)
                if error:
                    self.show_infobar(error, Gtk.MessageType.ERROR)
                elif self._abort_event and self._abort_event.is_set():
                    # TRANSLATORS: Notification shown in an infobar when
                    # chaff file generation is aborted by the user.
                    self.show_infobar(_("Chaff generation aborted"),
                                      Gtk.MessageType.WARNING)
                else:
                    # TRANSLATORS: Notification shown in an infobar when chaff file generation
                    # is complete.
                    self.show_infobar(_("Chaff generation complete"),
                                      Gtk.MessageType.INFO)

        def on_progress(fraction, msg=None, is_done=False, error=None):
            """Callback for progress bar"""
            # Use idle_add() because threads cannot make GDK calls.
            GLib.idle_add(_on_progress, fraction, msg, is_done, error)

        # TRANSLATORS: Progress message shown while generating chaff files.
        # 'Generating' is a present participle.
        # To indicate an ongoing operation, include the ellipsis as literal
        # Unicode (…) or as Unicode escape (\u2026).
        msg = _('Generating files\u2026')
        logger.info(msg)
        self.progressbar.show()
        self.progressbar.set_text(msg)
        self.progressbar.set_show_text(True)
        self.progressbar.set_fraction(0.0)
        self.make_button.set_sensitive(False)
        self.abort_button.set_sensitive(True)
        args = (stop_mode, stop_value, inspiration, output_dir,
                delete_when_finished, on_progress, self._abort_event)
        self.thread = threading.Thread(target=make_files_thread, args=args)
        self.thread.start()

    def run(self):
        """Run the dialog"""
        self.show_all()
        self.infobar.hide()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Code that is commonly shared throughout BleachBit
"""

import os
import re
import sys
from configparser import NoOptionError, RawConfigParser  # used in other files

from bleachbit import Log

APP_VERSION = "6.0.1"
APP_NAME = "BleachBit"
APP_URL = "https://www.bleachbit.org"
APP_COPYRIGHT = "Copyright (C) 2008-2026 Andrew Ziem"

socket_timeout = 10

if sys.version_info < (3, 8, 0):
    print('BleachBit requires Python version 3.8 or later')
    sys.exit(1)

if hasattr(sys, 'frozen'):
    stdout_encoding = 'utf-8'
else:
    stdout_encoding = sys.stdout.encoding

logger = Log.init_log()

# Setting below value to false disables update notification (useful
# for packages in repositories).
online_update_notification_enabled = True

#
# Platform
#

# platform
IS_WINDOWS = os.name == 'nt'
IS_POSIX = os.name == 'posix'
IS_LINUX = sys.platform.startswith('linux')
IS_MAC = sys.platform == 'darwin'
IS_BSD = sys.platform.startswith(('freebsd', 'openbsd', 'netbsd'))
IS_NETBSD = sys.platform[:6] == 'netbsd'
ARCH_BITS = 64 if sys.maxsize > 2**32 else 32

# file system attributes
FS_CASE_SENSITIVE = not (IS_WINDOWS or IS_MAC)
FS_SCAN_RE_FLAGS = 0 if FS_CASE_SENSITIVE else re.IGNORECASE


#
# Paths
#

# Windows
bleachbit_exe_path = None
if hasattr(sys, 'frozen'):
    # running frozen in py2exe
    bleachbit_exe_path = os.path.dirname(sys.executable)
    bleachbit_package_path = bleachbit_exe_path
else:
    # __file__ is absolute path to __init__.py
    bleachbit_package_path = os.path.dirname(__file__)
    bleachbit_exe_path = os.path.dirname(bleachbit_package_path)

# license
license_filename = None
license_filenames = ('/usr/share/common-licenses/GPL-3',  # Debian, Ubuntu
                     # Microsoft Windows
                     os.path.join(bleachbit_exe_path, 'COPYING'),
                     '/usr/share/doc/bleachbit-' + APP_VERSION + '/COPYING',  # CentOS, Fedora, RHEL
                     '/usr/share/licenses/bleachbit/COPYING',  # Fedora 21+, RHEL 7+
                     '/usr/share/doc/packages/bleachbit/COPYING',  # OpenSUSE 11.1
                     '/usr/pkg/share/doc/bleachbit/COPYING',  # NetBSD 5
                     '/usr/share/licenses/common/GPL3/license.txt')  # Arch Linux
for lf in license_filenames:
    if os.path.exists(lf):
        license_filename = lf
        break


def _home_dir():
    """Return home directory with fallback for missing HOME and passwd entry."""
    home = os.getenv('HOME')
    if home:
        return home
    # expanduser() falls back to a lookup in passwd database.
    home = os.path.expanduser('~')
    if home != '~':
        return home
    return '/tmp'


# configuration
portable_mode = False
options_dir = None
if IS_POSIX:
    options_dir = os.path.join(_home_dir(), ".config/bleachbit")
elif IS_WINDOWS:
    if os.path.exists(os.path.join(bleachbit_exe_path, 'bleachbit.ini')):
        # portable mode
        portable_mode = True
        options_dir = bleachbit_exe_path
    else:
        # installed mode
        options_dir = os.path.expandvars(r"${APPDATA}\BleachBit")

try:
    options_dir = os.environ['BLEACHBIT_TEST_OPTIONS_DIR']
except KeyError:
    pass

options_file = os.path.join(options_dir, "bleachbit.ini")

# check whether the application is running from the source tree
if not portable_mode:
    paths = (
        '../cleaners',
        '../Makefile',
        '../COPYING')
    existing = (
        os.path.exists(os.path.join(bleachbit_exe_path, path))
        for path in paths)
    portable_mode = all(existing)

# personal cleaners
personal_cleaners_dir = os.path.join(options_dir, "cleaners")

# system cleaners
# On Windows in portable mode, the bleachbit_exe_path is equal to
# options_dir, so be careful that system_cleaner_dir is not set to
# personal_cleaners_dir.
_exe_cleaners_dir = os.path.join(bleachbit_exe_path, 'cleaners')
_package_cleaners_dir = os.path.join(bleachbit_package_path, 'cleaners')
if os.path.isdir(_exe_cleaners_dir) and not portable_mode:
    system_cleaners_dir = _exe_cleaners_dir
elif os.path.isdir(_package_cleaners_dir) and not portable_mode:
    # AppImage
    system_cleaners_dir = _package_cleaners_dir
elif IS_LINUX or IS_MAC:
    system_cleaners_dir = '/usr/share/bleachbit/cleaners'
elif IS_WINDOWS:
    system_cleaners_dir = os.path.join(bleachbit_exe_path, 'share\\cleaners\\')
elif IS_NETBSD:
    system_cleaners_dir = '/usr/pkg/share/bleachbit/cleaners'
elif IS_BSD:
    system_cleaners_dir = '/usr/local/share/bleachbit/cleaners'
else:
    system_cleaners_dir = None
    logger.warning(
        'unknown system cleaners directory for platform %s ', sys.platform)

# local cleaners directory for running without installation (Windows or Linux)
local_cleaners_dir = None
if portable_mode:
    local_cleaners_dir = os.path.join(bleachbit_exe_path, 'cleaners')


def get_share_dirs():
    """Return ordered list of directories to search for shared data files."""
    if hasattr(sys, 'frozen'):
        # frozen in py2exe
        base_dirs = [
            os.path.join(bleachbit_exe_path, 'share'),
            bleachbit_exe_path,
        ]
    else:
        # installed .deb or .rpm has `__file__` = "/usr/share/bleachbit/__init__.py",
        # so that dirname() is "/usr/share/bleachbit"
        package_dir = bleachbit_package_path
        # When running from source, share directory is `../share/` from `__init__.py`.
        repo_root = os.path.normpath(os.path.join(package_dir, '..'))
        base_dirs = [
            os.path.join(package_dir, 'share'),
            os.path.join(repo_root, 'share')
        ]
    if system_cleaners_dir:
        # One directory up from the system cleaners directory.
        # This works when installed, like under `/usr/share`.
        base_dirs.append(os.path.dirname(system_cleaners_dir))
    # Remove duplicates while preserving the order.
    seen = set()
    unique_dirs = []
    for base_dir in base_dirs:
        if base_dir in seen:
            continue
        seen.add(base_dir)
        unique_dirs.append(base_dir)
    return unique_dirs


def get_share_path(filename):
    """Return path to a shared data file if it exists, else None."""
    for base_dir in get_share_dirs():
        candidate = os.path.normpath(os.path.join(base_dir, filename))
        if os.path.exists(candidate):
            return candidate
    logger.error('unknown location for %s', filename)
    return None


# windows10 theme
windows10_theme_path = os.path.normpath(
    os.path.join(bleachbit_exe_path, 'themes/windows10'))

# application icon
__icons = (
    # AppImage
    os.path.normpath(os.path.join(bleachbit_exe_path,
                                  'pixmaps/bleachbit.png')),
    # Linux
    '/usr/share/pixmaps/bleachbit.png',
    # NetBSD
    '/usr/pkg/share/pixmaps/bleachbit.png',
    # FreeBSD and OpenBSD
    os.path.normpath(os.path.join(bleachbit_exe_path,
                                  'share\\bleachbit.png')),  # Windows
    # When running from source (i.e., not installed).
    os.path.normpath(os.path.join(bleachbit_exe_path, 'bleachbit.png')),
)
appicon_path = None
for __icon in __icons:
    if os.path.exists(__icon):
        appicon_path = __icon

# locale directory
_exe_locale_dir = os.path.join(bleachbit_exe_path, 'locale')
if os.path.exists("./locale/"):
    # local locale (personal)
    locale_dir = os.path.abspath("./locale/")
elif os.path.exists(_exe_locale_dir):
    # AppImage
    locale_dir = _exe_locale_dir
# system-wide installed locale
elif sys.platform in ('linux', 'darwin'):
    locale_dir = "/usr/share/locale/"
elif sys.platform == "win32":
    locale_dir = os.path.join(bleachbit_exe_path, "share\\locale\\")
elif sys.platform[:6] == "netbsd":
    locale_dir = "/usr/pkg/share/locale/"
elif sys.platform.startswith("openbsd") or sys.platform.startswith("freebsd"):
    locale_dir = "/usr/local/share/locale/"


#
# URLs
#
base_url = "https://update.bleachbit.org"
help_contents_url = "https://www.bleachbit.org/help"
update_check_url = f"{base_url}/update/{APP_VERSION}"



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Import Winapp2.ini files
"""

import fnmatch
import glob
import logging
import os
import re
from xml.dom.minidom import parseString

import bleachbit
from bleachbit import Cleaner, Windows
from bleachbit.Action import Delete, Winreg
from bleachbit.Language import get_text as _

logger = logging.getLogger(__name__)


# TRANSLATORS: This is cleaner name for cleaners imported from winapp2.ini
langsecref_map = {
    '3001': ('winapp2_internet_explorer', 'Internet Explorer'),
    '3005': ('winapp2_edge_classic', 'Microsoft Edge'),
    '3006': ('winapp2_edge_chromium', 'Microsoft Edge'),
    # TRANSLATORS: This is cleaner name for cleaners imported from winapp2.ini
    '3021': ('winapp2_applications', _('Applications')),
    # TRANSLATORS: This is cleaner name for cleaners imported from winapp2.ini
    '3022': ('winapp2_internet', _('Internet')),
    # TRANSLATORS: This is cleaner name for cleaners imported from winapp2.ini
    '3023': ('winapp2_multimedia', _('Multimedia')),
    # TRANSLATORS: This is cleaner name for cleaners imported from winapp2.ini
    '3024': ('winapp2_utilities', _('Utilities')),
    '3025': ('winapp2_windows', 'Microsoft Windows'),
    '3026': ('winapp2_mozilla', 'Firefox/Mozilla'),
    '3027': ('winapp2_opera', 'Opera'),
    '3028': ('winapp2_safari', 'Safari'),
    '3029': ('winapp2_google_chrome', 'Google Chrome'),
    '3030': ('winapp2_thunderbird', 'Thunderbird'),
    '3031': ('winapp2_windows_store', 'Windows Store'),
    '3033': ('winapp2_vivaldi', 'Vivaldi'),
    '3034': ('winapp2_brave', 'Brave'),
    # Section=Games (technically not langsecref)
    'Games': ('winapp2_games', _('Games'))}


def xml_escape(s):
    """Lightweight way to escape XML entities"""
    return s.replace('&', '&amp;').replace('"', '&quot;').replace('<', '&lt;').replace('>', '&gt;')


def section2option(s):
    """Normalize section name to appropriate option name"""
    ret = re.sub(r'[^a-z0-9]', '_', s.lower())
    ret = re.sub(r'_+', '_', ret)
    ret = re.sub(r'(^_|_$)', '', ret)
    return ret


def _noop_progress(_fraction):
    """Default progress callback used when one is not provided."""
    return None


def _always_false():
    """Return False for cleaner auto_hide overrides."""
    return False


def detectos(required_ver, mock=False):
    """Returns boolean whether the detectos is compatible with the
    current operating system, or the mock version, if given."""
    # Do not compare as string because Windows 10 (build 10.0) comes after
    # Windows 8.1 (build 6.3).
    assert isinstance(required_ver, str)
    current_os = mock or Windows.parse_windows_build()
    required_ver = required_ver.strip()
    if '|' not in required_ver:
        # Exact version
        return Windows.parse_windows_build(required_ver) == current_os
    # Format of min|max
    req_min = required_ver.split('|')[0]
    req_max = required_ver.split('|')[1]
    if req_min and current_os < Windows.parse_windows_build(req_min):
        return False
    if req_max and current_os > Windows.parse_windows_build(req_max):
        return False
    return True


def winapp_expand_vars(pathname):
    """Expand environment variables using special Winapp2.ini rules"""
    # This is the regular expansion
    expand1 = os.path.expandvars(pathname)
    # Winapp2.ini expands %ProgramFiles% to %ProgramW6432%, etc.
    subs = (('ProgramFiles', 'ProgramW6432'),
            ('CommonProgramFiles', 'CommonProgramW6432'))
    for (sub_orig, sub_repl) in subs:
        pattern = re.compile(f'%{sub_orig}%', flags=re.IGNORECASE)
        if pattern.match(pathname):
            expand2 = pattern.sub(f'%{sub_repl}%', pathname)
            return expand1, os.path.expandvars(expand2)
    return expand1,


def detect_file(pathname):
    """Check whether a path exists for DetectFile#="""
    for expanded in winapp_expand_vars(pathname):
        for _i in glob.iglob(expanded):
            return True
    return False


def special_detect(code):
    """Check whether the SpecialDetect== software exists"""
    # The last two are used only for testing
    sd_keys = {'DET_CHROME': r'HKCU\Software\Google\Chrome',
               'DET_MOZILLA': r'HKCU\Software\Mozilla\Firefox',
               'DET_OPERA': r'HKCU\Software\Opera Software',
               'DET_THUNDERBIRD': r'HKLM\SOFTWARE\Clients\Mail\Mozilla Thunderbird',
               'DET_WINDOWS': r'HKCU\Software\Microsoft',
               'DET_SPACE_QUEST': r'HKCU\Software\Sierra Games\Space Quest'}
    if code in sd_keys:
        return Windows.detect_registry_key(sd_keys[code])
    else:
        logger.error('Unknown SpecialDetect=%s', code)
    return False


def fnmatch_translate(pattern):
    """Same as the original without the end"""
    ret = fnmatch.translate(pattern)
    if ret.endswith('$'):
        return ret[:-1]
    return re.sub(r'\\Z(\(\?ms\))?$', '', ret)


class Winapp:

    """Create cleaners from a Winapp2.ini-style file"""

    def __init__(self, pathname, cb_progress=_noop_progress):
        """Create cleaners from a Winapp2.ini-style file"""

        self.cleaners = {}
        self.cleaner_ids = []
        for langsecref in set(langsecref_map.values()):
            self.add_section(langsecref[0], langsecref[1])
        self.errors = 0
        self.parser = bleachbit.RawConfigParser()
        self.parser.read(pathname)
        self.re_detect = re.compile(r'^detect(\d+)?$')
        self.re_detectfile = re.compile(r'^detectfile(\d+)?$')
        self.re_excludekey = re.compile(r'^excludekey\d+$')
        section_total_count = len(self.parser.sections())
        section_done_count = 0
        for section in self.parser.sections():
            try:
                self.handle_section(section)
            except Exception:
                self.errors += 1
                logger.exception('parsing error in section %s', section)
            else:
                section_done_count += 1
                cb_progress(1.0 * section_done_count / section_total_count)

    def add_section(self, cleaner_id, name):
        """Add a section (cleaners)"""
        self.cleaner_ids.append(cleaner_id)
        self.cleaners[cleaner_id] = Cleaner.Cleaner()
        self.cleaners[cleaner_id].id = cleaner_id
        self.cleaners[cleaner_id].name = name
        assert name.strip() == name
        self.cleaners[cleaner_id].description = _('Imported from winapp2.ini')
        # The detect() function in this module effectively does what
        # auto_hide() does, so this avoids redundant, slow processing.
        self.cleaners[cleaner_id].auto_hide = _always_false

    def section_to_cleanerid(self, langsecref):
        """Given a langsecref (or section name), find the internal
        BleachBit cleaner ID."""
        # pre-defined, such as 3021
        if langsecref in langsecref_map.keys():
            return langsecref_map[langsecref][0]
        # custom, such as games
        cleanerid = 'winapp2_' + section2option(langsecref)
        if cleanerid not in self.cleaners:
            # never seen before
            self.add_section(cleanerid, langsecref)
        return cleanerid

    def excludekey_to_nwholeregex(self, excludekey):
        r"""Translate one ExcludeKey to CleanerML nwholeregex or return None for REG

        Supported examples
        FILE=%LocalAppData%\BleachBit\BleachBit.ini
        FILE=%LocalAppData%\BleachBit\|BleachBit.ini
        FILE=%LocalAppData%\BleachBit\|*.ini
        FILE=%LocalAppData%\BleachBit\|*.ini;*.bak
        PATH=%LocalAppData%\BleachBit\
        PATH=%LocalAppData%\BleachBit\|*.*
        REG|HKCU\Software\BleachBit
        """
        parts = excludekey.split('|')
        parts[0] = parts[0].upper()
        if parts[0] == 'REG':
            return None  # REG exclusions are handled separately

        # the last part contains the filename(s)
        files = None
        files_regex = ''
        if len(parts) == 3:
            files = parts[2].split(';')
            if len(files) == 1:
                # one file pattern like *.* or *.log
                files_regex = fnmatch_translate(files[0])
                if files_regex == '*.*':
                    files = None
            elif len(files) > 1:
                # multiple file patterns like *.log;*.bak
                files_regex = f"({'|'.join(fnmatch_translate(f) for f in files)})"

        # the middle part contains the file
        regexes = []
        for expanded in winapp_expand_vars(parts[1]):
            regex = None
            if not files:
                # There is no third part, so this is either just a folder,
                # or sometimes the file is specified directly.
                regex = fnmatch_translate(expanded)
            if files:
                # match one or more file types, directly in this tree or in any
                # sub folder
                regex = r'%s\\%s' % (
                    re.sub(r'\\\\((?:\))?)$', r'\1', fnmatch_translate(expanded)), files_regex)
            regexes.append(regex)

        if len(regexes) == 1:
            return regexes[0]
        else:
            return f"({'|'.join(regexes)})"

    def detect(self, section):
        """Check whether to show the section

        The logic:
        If the DetectOS does not match, the section is inactive.
        If any Detect or DetectFile matches, the section is active.
        If neither Detect or DetectFile was given, the section is active.
        Otherwise, the section is inactive.
        """
        if self.parser.has_option(section, 'detectos'):
            required_ver = self.parser.get(section, 'detectos')
            if not detectos(required_ver):
                return False
        any_detect_option = False
        if self.parser.has_option(section, 'specialdetect'):
            any_detect_option = True
            sd_code = self.parser.get(section, 'specialdetect')
            if special_detect(sd_code):
                return True
        for option in self.parser.options(section):
            if re.match(self.re_detect, option):
                # Detect= checks for a registry key
                any_detect_option = True
                key = self.parser.get(section, option)
                if Windows.detect_registry_key(key):
                    return True
            elif re.match(self.re_detectfile, option):
                # DetectFile= checks for a file
                any_detect_option = True
                key = self.parser.get(section, option)
                if detect_file(key):
                    return True
        return not any_detect_option

    def handle_section(self, section):
        """Parse a section"""
        # check whether the section is active (i.e., whether it will be shown)
        if not self.detect(section):
            return
        # excludekeys ignores a file, path, or registry key
        # FILE/PATH exclusions use regex patterns for file matching
        # REG exclusions are handled separately as registry key patterns
        file_excludekeys = []
        reg_excludekeys = []
        for option in self.parser.options(section):
            if re.match(self.re_excludekey, option):
                excludekey_val = self.parser.get(section, option)
                nwholeregex = self.excludekey_to_nwholeregex(excludekey_val)
                if nwholeregex is None:
                    # REG exclusion: extract the registry path
                    parts = excludekey_val.split('|')
                    if len(parts) >= 2:
                        reg_excludekeys.append(parts[1])
                else:
                    file_excludekeys.append(nwholeregex)
        # there are two ways to specify sections: langsecref= and section=
        if self.parser.has_option(section, 'langsecref'):
            # verify the langsecref number is known
            # langsecref_num is 3021, games, etc.
            langsecref_num = self.parser.get(section, 'langsecref')
        elif self.parser.has_option(section, 'section'):
            langsecref_num = self.parser.get(section, 'section')
        else:
            logger.error(
                'neither option LangSecRef nor Section found in section %s', section)
            return
        # find the BleachBit internal cleaner ID
        lid = self.section_to_cleanerid(langsecref_num)
        option_name = section.replace('*', '').strip()
        self.cleaners[lid].add_option(
            section2option(section), option_name, '')
        for option in self.parser.options(section):
            if (
                option
                in {
                    "default",
                    "langsecref",
                    "section",
                    "detectos",
                    "specialdetect",
                }
                or re.match(self.re_detect, option)
                or re.match(self.re_detectfile, option)
                or re.match(self.re_excludekey, option)
            ):
                continue
            if option.startswith('filekey'):
                self.handle_filekey(lid, section, option, file_excludekeys)
            elif option.startswith('regkey'):
                self.handle_regkey(lid, section, option, reg_excludekeys)
            elif option == 'warning':
                self.cleaners[lid].set_warning(
                    section2option(section), self.parser.get(section, 'warning'))
            else:
                logger.warning(
                    'unknown option %s in section %s', option, section)
                return

    def __make_file_provider(self, dirname, filename, recurse, removeself, excludekeys):
        """Change parsed FileKey to action provider"""
        regex = ''
        if recurse:
            search = 'walk.files'
            path = dirname
            if filename == '*.*':
                if removeself:
                    search = 'walk.all'
            else:
                regex = f' regex="^{xml_escape(fnmatch.translate(filename))}$" '
        else:
            search = 'glob'
            path = os.path.join(dirname, filename)
            if path.find('*') == -1:
                search = 'file'
        excludekeysxml = ''
        if excludekeys:
            if len(excludekeys) > 1:
                # multiple
                exclude_str = f"({'|'.join(excludekeys)})"
            else:
                # just one
                exclude_str = excludekeys[0]
            excludekeysxml = f'nwholeregex="{xml_escape(exclude_str)}"'
        action_str = f'<option command="delete" search="{search}" path="{xml_escape(path)}" {regex}{excludekeysxml}/>'
        yield Delete(parseString(action_str).childNodes[0])
        if removeself:
            search = 'file'
            if dirname.find('*') > -1:
                search = 'glob'
            action_str = f'<option command="delete" search="{search}" path="{xml_escape(dirname)}" type="d"/>'
            yield Delete(parseString(action_str).childNodes[0])

    def handle_filekey(self, lid, ini_section, ini_option, excludekeys):
        """Parse a FileKey# option.

        Section is [Application Name] and option is the FileKey#"""
        elements = self.parser.get(
            ini_section, ini_option).strip().split('|')
        dirnames = winapp_expand_vars(elements.pop(0))
        filenames = ""
        if elements:
            filenames = elements.pop(0)
        recurse = False
        removeself = False
        for element in elements:
            element = element.upper()
            if element == 'RECURSE':
                recurse = True
            elif element == 'REMOVESELF':
                recurse = True
                removeself = True
            else:
                logger.warning(
                    'unknown file option %s in section %s', element, ini_section)
        for filename in filenames.split(';'):
            for dirname in dirnames:
                # If dirname is a drive letter it needs a special treatment on Windows:
                # https://www.reddit.com/r/learnpython/comments/gawqne/why_cant_i_ospathjoin_on_a_drive_letterc/
                if os.path.splitdrive(dirname)[0] == dirname:
                    dirname = f'{dirname}{os.path.sep}'
                for provider in self.__make_file_provider(dirname, filename, recurse, removeself, excludekeys):
                    self.cleaners[lid].add_action(
                        section2option(ini_section), provider)

    def handle_regkey(self, lid, ini_section, ini_option, reg_excludekeys):
        """Parse a RegKey# option"""
        elements = self.parser.get(
            ini_section, ini_option).strip().split('|')
        path = elements[0]

        # Check if this registry key is excluded (exact match or starts with exclusion)
        for exclude_path in reg_excludekeys:
            # Normalize paths for comparison
            normalized_path = path.upper().replace('\\', '\\\\')
            normalized_exclude = exclude_path.upper().replace('\\', '\\\\')
            # Check if the path matches the exclusion (exact match or starts with exclusion)
            if normalized_path == normalized_exclude or \
               normalized_path.startswith(normalized_exclude + '\\\\'):
                logger.debug('Skipping excluded registry key: %s', path)
                return

        path = xml_escape(path)
        name = ""
        if len(elements) == 2:
            name = f' name="{xml_escape(elements[1])}"'
        action_str = f'<option command="winreg" path="{path}"{name}/>'
        provider = Winreg(parseString(action_str).childNodes[0])
        provider.excludekeys = reg_excludekeys
        self.cleaners[lid].add_action(section2option(ini_section), provider)

    def get_cleaners(self):
        """Return the created cleaners"""
        for cleaner_id in self.cleaner_ids:
            if self.cleaners[cleaner_id].is_usable():
                yield self.cleaners[cleaner_id]


def list_winapp_files():
    """List winapp2.ini files"""
    for dirname in (bleachbit.personal_cleaners_dir, bleachbit.system_cleaners_dir):
        fname = os.path.join(dirname, 'winapp2.ini')
        if os.path.exists(fname):
            yield fname


def load_cleaners(cb_progress=_noop_progress):
    """Scan for winapp2.ini files and load them"""
    cb_progress(0.0)
    for pathname in list_winapp_files():
        try:
            inicleaner = Winapp(pathname, cb_progress)
        except Exception:
            logger.exception(
                "Error reading winapp2.ini cleaner '%s'", pathname)
        else:
            for cleaner in inicleaner.get_cleaners():
                Cleaner.backends[cleaner.id] = cleaner
        yield True



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

# standard library
import json
import logging
import os
import threading
import time

# third party
from bleachbit.GtkShim import GLib, Gtk

# local import
import bleachbit
from bleachbit.Constant import URL_COOKIE_MGR
from bleachbit.Cookie import list_unique_cookies, load_keep_list, COOKIE_KEEP_LIST_FILENAME
from bleachbit.GuiBasic import open_url
from bleachbit.Language import get_text as _, nget_text as _n

logger = logging.getLogger(__name__)


COOKIE_DISCOVERY_WARN_THRESHOLD = 15.0  # seconds
COOKIE_DISCOVERY_DEBUG_THRESHOLD = 5.0  # seconds


class CookieManagerPane(Gtk.Box):
    """Widget for managing cookies to keep when cleaning."""

    def __init__(self, bottom_widget=None):
        Gtk.Box.__init__(
            self, orientation=Gtk.Orientation.VERTICAL, spacing=10)

        # Instructions label
        instructions = Gtk.Label()
        instructions.set_markup(
            # TRANSLATORS: Instruction label in the manage cookies dialog.
            "<b>" + _("Select the cookies to keep when cleaning cookies across browsers.") + "</b>")
        instructions.set_line_wrap(True)
        instructions.set_xalign(0)
        self.pack_start(instructions, False, False, 0)

        # Notice about supported cookie types
        notice_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        notice_label = Gtk.Label()
        notice_label.set_markup(
            # TRANSLATORS: This is a notice in the cookie manager about supported
            # cookie types. The hyperlink points to the documentation website.
            # Cookies refer to standard HTTP cookies.
            # Site data includes LocalStorage, IndexedDB, and other website storage.
            _("Only cookies are preserved. Site data is always deleted. "
              "<a href='{url}'>Learn more</a>").format(
                url=URL_COOKIE_MGR))
        notice_label.set_line_wrap(True)
        notice_label.set_xalign(0)
        notice_label.connect(
            "activate-link",
            lambda _label, url: open_url(url, parent_window=None, prompt=True))
        notice_box.pack_start(notice_label, False, False, 0)
        self.pack_start(notice_box, False, False, 0)

        # Search box
        search_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5)
        self.pack_start(search_box, False, False, 0)

        # TRANSLATORS: Label for the search entry in the manage cookies dialog.
        search_label = Gtk.Label(label=_("Search:"))
        search_box.pack_start(search_label, False, False, 0)

        self.search_entry = Gtk.Entry()
        # TRANSLATORS: Placeholder text in the search entry in the manage cookies dialog.
        # 'Filter' is a verb meaning to narrow down the list by typing.
        self.search_entry.set_placeholder_text(_("Filter cookies..."))
        self.search_entry.connect("changed", self.on_search_changed)
        search_box.pack_start(self.search_entry, True, True, 0)

        # TRANSLATORS: Toggle button label in the manage cookies dialog.
        # "Show" is a command (imperative).
        # 'Selected' refers to selected cookies (the noun "cookies" is implied).
        self.selected_toggle = Gtk.ToggleButton(label=_("Show Selected"))
        self.selected_toggle.set_tooltip_text(
            # TRANSLATORS: Tooltip for the "Show Selected" toggle button.
            _("Only show cookies that are currently selected"))
        self.selected_toggle.connect("toggled", self.on_selected_toggle)
        search_box.pack_start(self.selected_toggle, False, False, 0)

        self.show_selected_only = False

        # Create scrollable window for cookie list
        scrolled = Gtk.ScrolledWindow()
        scrolled.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        self.pack_start(scrolled, True, True, 0)

        self.keep_list_path = os.path.join(
            bleachbit.options_dir, COOKIE_KEEP_LIST_FILENAME)
        self.saved_domains = load_keep_list()
        self._is_loading = False

        # Create cookie list store: checkbox, domain
        self.cookie_store = Gtk.ListStore(bool, str)

        # Create filter for the list store
        self.cookie_filter = self.cookie_store.filter_new()
        self.cookie_filter.set_visible_func(self.filter_cookies)

        # Create the TreeView
        self.treeview = Gtk.TreeView(model=self.cookie_filter)

        # Create columns
        renderer_toggle = Gtk.CellRendererToggle()
        renderer_toggle.connect("toggled", self.on_cell_toggled)
        column_toggle = Gtk.TreeViewColumn("", renderer_toggle, active=0)
        self.treeview.append_column(column_toggle)

        renderer_text = Gtk.CellRendererText()
        # TRANSLATORS: Column header in the manage cookies dialog.
        # 'Host' is a noun meaning the website hostname (domain name) of the cookie.
        column_domain = Gtk.TreeViewColumn(_("Host"), renderer_text, text=1)
        column_domain.set_sort_column_id(1)
        column_domain.set_resizable(True)
        column_domain.set_expand(True)
        self.treeview.append_column(column_domain)

        scrolled.add(self.treeview)

        # Button box
        button_box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        button_box.set_halign(Gtk.Align.END)
        self.pack_start(button_box, False, False, 0)

        # Stat label
        self.stat_label = Gtk.Label()
        self.update_stat_label()
        button_box.pack_start(self.stat_label, True, True, 0)

        # Select all / Deselect all buttons
        # TRANSLATORS: Button label in the manage cookies dialog.
        self.select_all_btn = Gtk.Button.new_with_label(_("Select All"))
        self.select_all_btn.connect("clicked", self.on_select_all_clicked)
        button_box.pack_start(self.select_all_btn, False, False, 0)

        # TRANSLATORS: Button label in the manage cookies dialog.
        self.deselect_all_btn = Gtk.Button.new_with_label(_("Deselect All"))
        self.deselect_all_btn.connect("clicked", self.on_deselect_all_clicked)
        button_box.pack_start(self.deselect_all_btn, False, False, 0)

        if bottom_widget is not None:
            self.pack_start(bottom_widget, False, False, 0)

        self._populate_cookie_store()

    def save_changes(self):
        """Save choices to a JSON file."""
        keep_list = sorted(self._iter_selected_domains())

        try:
            with open(self.keep_list_path, "w", encoding="utf-8") as f:
                json.dump(keep_list, f, indent=2)
            self.saved_domains = set(keep_list)
            return True
        except OSError as exc:
            logger.error("Failed to save cookie keep list %s: %s",
                         self.keep_list_path, exc)
            return False

    def update_stat_label(self):
        """Update the stat label: how many selected"""
        total = len(self.cookie_store)
        selected = sum(1 for row in self.cookie_store if row[0])
        visible = sum(1 for _row in self.cookie_filter)
        if visible < total:
            # TRANSLATORS: %(selected)d is the count of selected cookies,
            # %(total)d is the total count, %(visible)d is the visible count
            self.stat_label.set_text(
                _n("%(selected)d of %(total)d cookie kept (%(visible)d visible)",
                   "%(selected)d of %(total)d cookies kept (%(visible)d visible)",
                   selected) % {'selected': selected, 'total': total, 'visible': visible})
        else:
            # TRANSLATORS: %(selected)d is the count of selected cookies,
            # %(total)d is the total count
            self.stat_label.set_text(
                _n("%(selected)d of %(total)d cookie kept",
                   "%(selected)d of %(total)d cookies kept",
                   selected) % {'selected': selected, 'total': total})

    def on_cell_toggled(self, _widget, path):
        """Toggle the checkbox in the child model"""
        # Convert path from filter model to child model
        filter_path = Gtk.TreePath.new_from_string(path)
        child_path = self.cookie_filter.convert_path_to_child_path(filter_path)

        # Toggle the checkbox in the child model
        self.cookie_store[child_path][0] = not self.cookie_store[child_path][0]
        self.save_changes()
        self.update_stat_label()

    def on_select_all_clicked(self, _widget):
        """Select all cookies"""
        if self._is_loading:
            return
        self._set_filtered_selection(True)
        self.save_changes()
        self.update_stat_label()

    def on_deselect_all_clicked(self, _widget):
        """Deselect all cookies"""
        if self._is_loading:
            return
        self._set_filtered_selection(False)
        self.save_changes()
        self.update_stat_label()

    def _set_filtered_selection(self, is_selected):
        """Set selection state only for rows visible in the current filter."""
        # Collect paths first to prevent iterator invalidation when rows disappear
        # from the filter (e.g., deselecting while 'Show Selected' is active).
        paths = []
        tree_iter = self.cookie_filter.get_iter_first()
        while tree_iter:
            child_iter = self.cookie_filter.convert_iter_to_child_iter(
                tree_iter)
            if child_iter:
                paths.append(self.cookie_store.get_path(child_iter))
            tree_iter = self.cookie_filter.iter_next(tree_iter)

        for path in paths:
            self.cookie_store[path][0] = is_selected

    def filter_cookies(self, model, tree_iter, _data):
        """Filter function for the cookie list"""
        search_text = self.search_entry.get_text().lower()
        if not search_text:
            matches_search = True
        else:
            domain = model[tree_iter][1].lower()
            matches_search = search_text in domain

        if not matches_search:
            return False

        if self.show_selected_only:
            return bool(model[tree_iter][0])

        return True

    def on_selected_toggle(self, widget):
        """Toggle whether only selected cookies should be visible."""
        self.show_selected_only = widget.get_active()
        self.cookie_filter.refilter()
        self.update_stat_label()

    def on_search_changed(self, _widget):
        """Called when the search text changes"""
        self.cookie_filter.refilter()
        self.update_stat_label()

    def _populate_cookie_store(self):
        """Populate the list store with discovered and saved cookie hosts."""
        self._is_loading = True
        self._spinner = Gtk.Spinner()
        self._spinner.set_size_request(24, 24)
        self._spinner.start()
        # Insert spinner row placeholder so the button_box area is not empty
        # while loading; we embed the spinner above the treeview instead.
        self.treeview.set_sensitive(False)
        self.select_all_btn.set_sensitive(False)
        self.deselect_all_btn.set_sensitive(False)
        # Find the scrolled window's parent vbox and insert a spinner overlay
        scrolled = self.treeview.get_parent()
        vbox = scrolled.get_parent()
        spinner_box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        spinner_box.set_halign(Gtk.Align.CENTER)
        spinner_box.pack_start(self._spinner, False, False, 0)
        # TRANSLATORS: Status message shown while the cookie list is loading
        # in the cookie manager dialog window.
        # "Loading" is a present participle (ongoing action).
        # To indicate an ongoing operation, include the ellipsis as literal
        # Unicode (…) or as Unicode escape (\u2026).
        self._loading_label = Gtk.Label(label=_("Loading cookies\u2026"))
        spinner_box.pack_start(self._loading_label, False, False, 0)
        vbox.pack_start(spinner_box, False, False, 0)
        vbox.reorder_child(spinner_box, vbox.get_children().index(scrolled))
        self._spinner_box = spinner_box
        self.show_all()

        def _worker():
            start = time.monotonic()
            discovered = []
            try:
                discovered = list_unique_cookies()
            except (OSError, RuntimeError, ValueError) as exc:  # pragma: no cover - defensive logging
                logger.error("Failed to enumerate cookies: %s", exc)
            duration = time.monotonic() - start
            if duration >= COOKIE_DISCOVERY_WARN_THRESHOLD:
                logger.warning("Enumerating cookie hosts took %.2fs", duration)
            elif duration >= COOKIE_DISCOVERY_DEBUG_THRESHOLD:
                logger.debug("Enumerating cookie hosts took %.2fs", duration)
            GLib.idle_add(self._finish_populate, discovered)

        t = threading.Thread(target=_worker, daemon=True)
        t.start()

    def _finish_populate(self, discovered):
        """Called on the GTK main thread after cookie discovery finishes."""
        self._is_loading = False
        self._spinner.stop()
        self._spinner_box.destroy()
        self.treeview.set_sensitive(True)
        self.select_all_btn.set_sensitive(True)
        self.deselect_all_btn.set_sensitive(True)

        all_hosts = {h.strip() for h in discovered if h}
        all_hosts.update(self.saved_domains)
        sorted_hosts = sorted(all_hosts, key=lambda host: host.lower())

        for host in sorted_hosts:
            is_saved = host in self.saved_domains
            self.cookie_store.append([is_saved, host])

        self.update_stat_label()
        return False

    def _iter_selected_domains(self):
        """Yield domain names for all selected (checked) cookies."""
        for row in self.cookie_store:
            if row[0]:
                yield row[1]



# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Logging
"""

import logging
import sys


def is_debugging_enabled_via_cli():
    """Return boolean whether user required debugging on the command line"""
    if 'unittest' in sys.modules:
        return True
    return any(arg.startswith('--debug') for arg in sys.argv)


class DelayLog(object):
    def __init__(self):
        self.queue = []
        self.msg = ''

    def read(self):
        yield from self.queue
        self.queue = []

    def write(self, msg):
        self.msg += msg
        if self.msg[-1] == '\n':
            self.queue.append(self.msg)
            self.msg = ''


def init_log():
    """Set up the root logger

    This is one of the first steps in __init___
    """
    logger = logging.getLogger('bleachbit')
    # On Microsoft Windows when running frozen without the console,
    # avoid py2exe redirecting stderr to bleachbit.exe.log by not
    # writing to stderr because py2exe redirects stderr to a file.
    #
    # sys.frozen = 'console_exe' means the console is shown, which
    # does not require special handling.
    if hasattr(sys, 'frozen') and sys.frozen == 'windows_exe':  # pylint: disable=no-member
        sys.stderr = DelayLog()

    # debug if command line asks for it or if this a non-final release
    if is_debugging_enabled_via_cli():
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    logger_sh = logging.StreamHandler()
    console_formatter = logging.Formatter('%(message)s')
    logger_sh.setFormatter(console_formatter)
    logger.addHandler(logger_sh)

    # If --debug-log parameter was passed, set up the file handler here instead
    # of in CLI.py, so logs are captured from the very beginning.
    debug_log_path = None
    for i, arg in enumerate(sys.argv):
        # --debug-log /path/file format (space delimited)
        if arg == '--debug-log' and i + 1 < len(sys.argv):
            debug_log_path = sys.argv[i + 1]
            break
        # --debug-log=/path/file format (delimited with equals sign)
        if arg.startswith('--debug-log='):
            debug_log_path = arg.split('=', 1)[1]
            break

    if debug_log_path:
        file_handler = logging.FileHandler(debug_log_path)
        # Always use DEBUG level for log file.
        file_handler.setLevel(logging.DEBUG)
        # removed: %(name)s
        file_formatter = logging.Formatter(
            '%(asctime)s -  %(levelname)s - %(message)s')
        file_handler.setFormatter(file_formatter)
        logger.addHandler(file_handler)
        logger.debug('Debug log file initialized at %s', debug_log_path)

    return logger


def set_root_log_level(is_debug=False):
    """Adjust the root log level

    This runs later in the application's startup process when the
    configuration is loaded or after a change via the GUI.
    """
    root_logger = logging.getLogger('bleachbit')
    is_debug_effective = is_debug or is_debugging_enabled_via_cli()
    root_logger.setLevel(logging.DEBUG if is_debug_effective else logging.INFO)


class GtkLoggerHandler(logging.Handler):
    def __init__(self, append_text):
        logging.Handler.__init__(self)
        self.append_text = append_text
        self.msg = ''
        self.update_log_level()

    def update_log_level(self):
        """Set the log level"""
        from bleachbit.Options import options
        if is_debugging_enabled_via_cli() or options.get('debug'):
            self.min_level = logging.DEBUG
        else:
            self.min_level = logging.WARNING

    def emit(self, record):
        if record.levelno < self.min_level:
            return
        tag = 'error' if record.levelno >= logging.WARNING else None
        msg = record.getMessage()
        if record.exc_text:
            msg = msg + '\n' + record.exc_text
        self.append_text(msg + '\n', tag)

    def write(self, msg):
        self.msg += msg
        if self.msg[-1] == '\n':
            tag = None
            self.append_text(msg, tag)
            self.msg = ''



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
WindowInfo class and utility functions for GUI
"""

import os
import threading
from enum import Enum
from typing import Optional

from bleachbit import APP_NAME, FileUtilities, IS_WINDOWS
from bleachbit.GUI import logger
from bleachbit.GtkShim import (
    GLib, Gdk, Gtk, gi,
    suppress_pygobject_asyncio_warnings,
)


class WindowInfo:
    def __init__(self, x, y, width, height, monitor_model):
        super().__init__()
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.monitor_model = monitor_model

    def __str__(self):
        return f"WindowInfo(x={self.x}, y={self.y}, width={self.width}, height={self.height}, monitor_model={self.monitor_model})"


def clear_clipboard():
    """Clear the clipboard buffer"""
    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    clipboard.set_text(' ', 1)
    clipboard.clear()
    flush_gtk_events()
    # GTK may leave the clipboard locked, so the win32api may
    # get an "access denied" error.
    if IS_WINDOWS:
        import bleachbit.Windows  # pylint: disable=import-outside-toplevel
        try:
            bleachbit.Windows.clear_clipboard()
        except bleachbit.Windows.pywintypes.error as e:
            winerror = getattr(e, 'winerror', e.args[0] if e.args else None)
            if winerror != 5:
                raise
            logger.debug(
                'Failed to clear Windows clipboard using win32 API',
                exc_info=True)


def get_clipboard_paths(clipboard=None, targets=None):
    """Returns paths from clipboard as a list"""
    if clipboard is None:
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    if targets is None:
        has_targets, targets = clipboard.wait_for_targets()
        if not has_targets:
            targets = []

    shred_paths = []
    if IS_WINDOWS and Gdk.atom_intern_static_string('FileNameW') in targets:
        # Windows
        # Use non-GTK+ functions because because GTK+ 2 does not work.
        import bleachbit.Windows  # pylint: disable=import-outside-toplevel
        shred_paths = bleachbit.Windows.get_clipboard_paths()
    elif Gdk.atom_intern_static_string('text/uri-list') in targets:
        # Linux
        shred_uri_contents = clipboard.wait_for_contents(
            Gdk.atom_intern_static_string('text/uri-list'))
        if shred_uri_contents:
            shred_paths = FileUtilities.uris_to_paths(
                shred_uri_contents.get_uris())

    if not shred_paths and (
            Gdk.atom_intern_static_string('text/plain') in targets or
            Gdk.atom_intern_static_string('UTF8_STRING') in targets):
        # Plain text pasted from a text editor
        text = clipboard.wait_for_text()
        if text:
            shred_paths = [p.strip()
                           for p in text.splitlines() if p.strip()]
    return shred_paths


def get_font_size_from_name(font_name):
    """Get the font size from the font name"""
    if not isinstance(font_name, str):
        return None
    if not font_name:
        return None
    try:
        number_part = font_name.split()[-1]
    except IndexError:
        return None
    if '.' in number_part:
        return int(float(number_part))
    try:
        size_int = int(number_part)
    except ValueError:
        return None
    if size_int < 1:
        return None
    return size_int


def get_window_info(window):
    """Get the geometry and monitor of a window.

    window: Gtk.Window

    https://docs.gtk.org/gdk3/method.Screen.get_monitor_at_window.html
    Deprecated since: 3.22
    Use gdk_display_get_monitor_at_window() instead.

    https://docs.gtk.org/gdk3/method.Display.get_monitor_at_window.html
    Available since: 3.22

    https://docs.gtk.org/gdk3/method.Screen.get_monitor_geometry.html
    Deprecated since: 3.22
    Use gdk_monitor_get_geometry() instead.

    https://docs.gtk.org/gdk3/method.Monitor.get_geometry.html
    Available since: 3.22

    Returns a Rectangle-like object with with extra `monitor_model`
    property with the monitor model string.
    """
    assert window is not None
    assert isinstance(window, Gtk.Window)
    gdk_window = window.get_window()
    display = Gdk.Display.get_default()
    assert display is not None
    monitor = display.get_monitor_at_window(gdk_window)
    assert monitor is not None
    geo = monitor.get_geometry()
    assert geo is not None
    assert isinstance(geo, Gdk.Rectangle)
    if display.get_n_monitors() > 0 and monitor.get_model():
        monitor_model = monitor.get_model()
    else:
        monitor_model = "(unknown)"
    return WindowInfo(geo.x, geo.y, geo.width, geo.height, monitor_model)


class ThemeChangeStatus(Enum):
    """State machine for detecting whether a theme update occurred."""

    CHANGED = "changed"
    UNCHANGED = "unchanged"
    UNKNOWN = "unknown"


def detect_dark_background(widget: Optional[Gtk.Widget]) -> Optional[bool]:
    """Return True if the widget background is dark, False if light, None on failure."""
    threshold = 0.45
    if widget is None:
        return None

    try:
        style_context = widget.get_style_context()
        rgba = None
        if style_context is not None and hasattr(style_context, 'lookup_color'):
            lookup = style_context.lookup_color('theme_bg_color')
            if lookup:
                rgba = lookup[-1] if isinstance(lookup, tuple) else lookup

        if rgba is None:
            return None

        luminance = 0.2126 * rgba.red + 0.7152 * rgba.green + 0.0722 * rgba.blue
        is_dark = luminance < threshold
        # logger.debug("Detected widget luminance=%f -> dark=%s", luminance, is_dark)
        return is_dark
    except Exception:
        logger.debug("Failed to detect widget background", exc_info=True)
        return None


def classify_theme_change(before_dark: Optional[bool], after_dark: Optional[bool]) -> ThemeChangeStatus:
    """Compare observations before and after a toggle to classify change."""
    if before_dark is None or after_dark is None:
        return ThemeChangeStatus.UNKNOWN
    if before_dark == after_dark:
        return ThemeChangeStatus.UNCHANGED
    return ThemeChangeStatus.CHANGED


def should_show_dark_mode_warning(before_dark: Optional[bool], after_dark: Optional[bool]) -> bool:
    """Return True when we should warn the user about theme toggles."""
    status = classify_theme_change(before_dark, after_dark)
    # Warn when the theme did not change or when we cannot tell (UNKNOWN).
    return status != ThemeChangeStatus.CHANGED


def flush_gtk_events(max_iterations: int = 5):
    """Process pending GTK events to allow style updates to land."""
    iterations = 0
    while Gtk.events_pending() and (max_iterations is None or iterations < max_iterations):
        # PyGObject 3.56.2 calls deprecated asyncio APIs, which breaks tests
        # run with PYTHONWARNINGS=error.
        with suppress_pygobject_asyncio_warnings():
            Gtk.main_iteration_do(False)
        iterations += 1


def notify(msg):
    """Show a popup-notification"""
    import importlib
    if importlib.util.find_spec('plyer'):
        # On Windows, use Plyer.
        notify_plyer(msg)
        return
    # On Linux, use GTK Notify.
    notify_gi(msg)


def notify_gi(msg):
    """Show a pop-up notification.

    The Windows pygy-aio installer does not include notify, so this is just for Linux.
    """
    try:
        gi.require_version('Notify', '0.7')
    except ValueError as e:
        logger.debug('gi.require_version("Notify", "0.7") failed: %s', e)
        return
    from gi.repository import Notify
    if Notify.init(APP_NAME):
        notification_obj = Notify.Notification.new(
            'BleachBit', msg, 'bleachbit')
        notification_obj.set_hint(
            "desktop-entry", GLib.Variant('s', 'bleachbit'))
        try:
            notification_obj.show()
        except gi.repository.GLib.GError as e:
            logger.debug('Notify.Notification.show() failed: %s', e)
            return
        notification_obj.set_timeout(10000)


def notify_plyer(msg):
    """Show a pop-up notification.

    Linux distributions do not include plyer, so this is just for Windows.
    """
    if not os.name == 'nt':
        raise RuntimeError("notify_plyer() is only for Windows")
    from bleachbit import bleachbit_exe_path

    # On Windows 10,  PNG does not work.
    __icon_fns = (
        os.path.normpath(os.path.join(bleachbit_exe_path,
                                      'share\\bleachbit.ico')),
        os.path.normpath(os.path.join(bleachbit_exe_path,
                                      'windows\\bleachbit.ico')))

    icon_fn = None
    for __icon_fn in __icon_fns:
        if os.path.exists(__icon_fn):
            icon_fn = __icon_fn
            break

    from plyer import notification
    notification.notify(
        title=APP_NAME,
        message=msg,
        app_name=APP_NAME,  # not shown on Windows 10
        app_icon=icon_fn,
    )


def threaded(func):
    """Decoration to create a threaded function"""
    def wrapper(*args):
        thread = threading.Thread(target=func, args=args)
        thread.start()
    return wrapper



# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Check for updates via the Internet
"""

# standard library
import hashlib
import logging
import os
import socket
import struct
import sys
import platform
from collections.abc import Callable
try:
    from urllib3.util.retry import Retry
    HAVE_URLLIB3 = True
except ImportError:
    HAVE_URLLIB3 = False

# third party
try:
    import requests
    HAVE_REQUESTS = True
except ImportError:
    HAVE_REQUESTS = False

if HAVE_REQUESTS:
    RequestException = requests.exceptions.RequestException
else:
    class RequestException(Exception):
        pass

# local imports
from bleachbit import bleachbit_exe_path, APP_VERSION
from bleachbit.FileUtilities import delete
from bleachbit.Language import get_active_language_code, get_text as _

logger = logging.getLogger(__name__)


def unset_sslkeylogfile(use_logger):
    """Unset environment variable SSLKEYLOGFILE

    Workaround for an OpenSSL crash before checking for updates.
    https://github.com/bleachbit/bleachbit/issues/1826

    Returns True if unset
    """
    if not os.name == 'nt':
        return False
    if not os.environ.get('SSLKEYLOGFILE'):
        return False
    del os.environ['SSLKEYLOGFILE']
    if use_logger:
        logger.debug('The environment variable SSLKEYLOGFILE is not supported')
    return True


def download_url_to_fn(url, fn, expected_sha512=None, on_error=None,
                       max_retries=3, backoff_factor=0.5, timeout=60):
    """Download a URL to the given filename

    fn: target filename

    expected_sha512: expected SHA-512 hash

    on_error: callback function in case of error

    max_retries: retry count

    backoff_factor: how long to wait before retries

    timeout: number of seconds to wait to establish connection

    return: True if succeeded, False if failed
    """
    logger.info('Downloading %s to %s', url, fn)
    assert isinstance(url, str)
    assert isinstance(fn, str), f'fn is not a string: {repr(fn)}'
    assert isinstance(expected_sha512, (type(None), str))
    assert isinstance(on_error, (type(None), Callable))
    assert isinstance(max_retries, int)
    assert isinstance(backoff_factor, float)
    assert isinstance(timeout, int)

    unset_sslkeylogfile(True)

    msg = _('Downloading URL failed: %s') % url

    def do_error(msg2):
        if on_error:
            on_error(msg, msg2)
        delete(fn, ignore_missing=True)  # delete any partial download

    try:
        response = fetch_url(url)
    except RequestException as exc:
        # For retryable errors (like 503), use a simplified error message
        if HAVE_REQUESTS and isinstance(exc, requests.exceptions.RetryError):
            msg2 = 'Server temporarily unavailable (retries exceeded)'
            logger.warning("%s: %s", msg, type(exc).__name__)
        else:
            msg2 = f'{type(exc).__name__}: {exc}'
            logger.exception(msg)
        do_error(msg2)
        return False
    if response.status_code != 200:
        logger.error(msg)
        msg2 = f'HTTP status code: {response.status_code}'
        do_error(msg2)
        return False
    if expected_sha512:
        hash_actual = hashlib.sha512(response.content).hexdigest()
        if hash_actual != expected_sha512:
            msg2 = f"SHA-512 mismatch: expected {expected_sha512}, got {hash_actual}"
            do_error(msg2)
            return False
    fn_dir = os.path.dirname(fn)
    if not os.path.exists(fn_dir):
        os.makedirs(fn_dir)
    with open(fn, 'wb') as f:
        f.write(response.content)
    return True


def fetch_url(url, max_retries=3, backoff_factor=0.5, timeout=60,
              headers=None):
    """Fetch a URL using requests library

    Args:
        url (str): URL to fetch content from
        max_retries (int, optional): Maximum number of retry attempts
        backoff_factor (float, optional): A backoff factor to apply between attempts
        timeout (int, optional): How many seconds to wait for the server before giving up
        headers (dict, optional): Extra HTTP headers appended to default headers

    Returns:
        requests.Response: Response object from requests library

    Raises:
        requests.RequestException: If there is an error fetching the URL
    """
    assert isinstance(url, str)
    assert url.startswith('http'), f"URL must start with http, got {url}"
    assert isinstance(max_retries, int)
    assert max_retries >= 0
    assert isinstance(backoff_factor, float)
    assert backoff_factor > 0
    assert isinstance(timeout, int)
    assert timeout > 0
    if hasattr(sys, 'frozen'):
        # when frozen by py2exe, certificates are in alternate location
        ca_bundle = os.path.join(bleachbit_exe_path, 'cacert.pem')
        if os.path.exists(ca_bundle):
            requests.utils.DEFAULT_CA_BUNDLE_PATH = ca_bundle
            requests.adapters.DEFAULT_CA_BUNDLE_PATH = ca_bundle
        else:
            logger.error(
                'Application is frozen but certificate file not found: %s', ca_bundle)
    if not HAVE_REQUESTS:
        raise RequestException(
            'The requests package is not installed: network features are disabled.')
    assert headers is None or isinstance(headers, dict)
    request_headers = {'User-Agent': get_user_agent()}
    if headers:
        request_headers.update(headers)
    unset_sslkeylogfile(True)
    # 408: request timeout
    # 429: too many requests
    # 500: internal server error
    # 502: bad gateway
    # 503: service unavailable
    # 504: gateway_timeout
    status_forcelist = (408, 429, 500, 502, 503, 504)
    with requests.Session() as session:
        if HAVE_URLLIB3:
            retries = Retry(total=max_retries, backoff_factor=backoff_factor,
                            status_forcelist=status_forcelist, redirect=5)
            session.mount(
                'http://', requests.adapters.HTTPAdapter(max_retries=retries))
            session.mount(
                'https://', requests.adapters.HTTPAdapter(max_retries=retries))
        response = session.get(url, headers=request_headers,
                               timeout=timeout, verify=True)
    return response


def _get_os_name_version():
    """Return (os_name, os_version) tuple for network requests."""
    os_name = platform.system()  # 'Linux', 'Windows', etc.
    if sys.platform == 'linux':
        # pylint: disable=import-outside-toplevel
        from bleachbit.Unix import get_distribution_name_version
        os_version = get_distribution_name_version()
    elif sys.platform[:6] == 'netbsd':
        os_version = os_name + '/' + platform.machine() + ' ' + platform.release()
    else:
        os_version = platform.uname().version
    return os_name, os_version


def get_update_request_headers():
    """Return headers specific to update checks."""
    os_name, os_version = _get_os_name_version()

    headers = {
        'X-BleachBit-Version': APP_VERSION,
        'X-OS-Type': os_name,
        'X-OS-Version': os_version,
        'X-Locale': get_active_language_code(),
    }

    if (gtk_version := get_gtk_version()):
        headers['X-GTK-Version'] = gtk_version

    if os.name == 'nt':
        headers['X-Python-Version'] = platform.python_version()
        headers['X-Pointer-Bits'] = str(8 * struct.calcsize('P'))

    return headers


def get_gtk_version():
    """Return the version of GTK

    If GTK is not available, returns None.
    """
    # pylint: disable=import-outside-toplevel, import-error
    from bleachbit.GtkShim import Gtk, HAVE_GTK
    if not HAVE_GTK:
        return None
    gtk_version = (Gtk.get_major_version(),
                   Gtk.get_minor_version(), Gtk.get_micro_version())
    return '.'.join([str(x) for x in gtk_version])


def get_ip_for_url(url):
    """Given an https URL, return the IP address"""
    if not url:
        return '(no URL)'
    url_split = url.split('/')
    if len(url_split) < 3:
        return '(bad URL)'
    hostname = url.split('/')[2]
    try:
        ip_address = socket.gethostbyname(hostname)
    except socket.gaierror:
        return '(socket.gaierror)'
    return ip_address


def get_user_agent():
    """Return the user agent string"""
    os_name, os_ver = _get_os_name_version()
    locale = get_active_language_code()
    parts = [os_name, os_ver, locale]
    if (gtk_ver := get_gtk_version()):
        parts.append(f'GTK {gtk_ver}')

    agent = f"BleachBit/{APP_VERSION} ({'; '.join(parts)})"
    return agent



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Enumerate and terminate processes
"""

import signal
import glob
import subprocess
from collections import namedtuple
import os

from bleachbit import IS_LINUX, IS_POSIX, IS_WINDOWS

ProcessInfo = namedtuple('ProcessInfo', ['pid', 'name', 'same_user'])

try:
    import psutil
    _has_psutil = True
except ImportError:
    _has_psutil = False


def enumerate_processes():
    """Yield ProcessInfo(pid, name, same_user) for all accessible processes.

    'same_user' is True if the process owner matches the current (real) user.
    On Unix with sudo, compares against the non-root user.
    """
    if _has_psutil and IS_POSIX:
        yield from _enumerate_psutil_posix()
        return
    # Windows should always have psutil.
    if _has_psutil and IS_WINDOWS:
        yield from _enumerate_psutil_windows()
        return
    if IS_LINUX:
        yield from _enumerate_proc_fs()
        return
    if IS_POSIX:
        yield from _enumerate_ps_aux()
        return
    raise RuntimeError('no method to enumerate processes on this system')


def _enumerate_psutil_posix():
    """Enumerate processes with psutils on POSIX"""
    from bleachbit.General import get_real_uid
    target_uid = get_real_uid()
    for proc in psutil.process_iter(['name', 'exe', 'uids']):
        try:
            name = proc.info['name']
            if not name:
                continue
            exe = proc.info.get('exe')
            uids = proc.info.get('uids')
            same_user = uids is not None and uids.real == target_uid
            yield ProcessInfo(proc.pid, name, same_user)
            # exe basename may differ from name (e.g. truncated comm)
            if exe and os.path.basename(exe) != name:
                yield ProcessInfo(proc.pid, os.path.basename(exe), same_user)
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            continue


def _enumerate_psutil_windows():
    """Enumerate processes with psutils on Windows"""

    current_user = psutil.Process().username().lower()
    for proc in psutil.process_iter(['name', 'username']):
        try:
            name = proc.info['name']
            if not name:
                continue
            same_user = (proc.info['username'] or '').lower() == current_user
            yield ProcessInfo(proc.pid, name, same_user)
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            continue


def _enumerate_proc_fs():
    """/proc filesystem strategy (Linux)"""
    from bleachbit.General import get_real_uid
    target_uid = get_real_uid()
    for filename in glob.iglob("/proc/*/exe"):
        pid_dir = os.path.dirname(filename)
        pid = int(os.path.basename(pid_dir))
        name = None
        try:
            target = os.path.realpath(filename)
            # Google Chrome 74 on Ubuntu 19.04 showed up as
            # /opt/google/chrome/chrome (deleted)
            name = os.path.basename(target).replace(' (deleted)', '')
        except OSError:
            # 13 = permission denied
            pass
        except TypeError:
            # TypeError happens, for example, when link points to
            # '/etc/password\x00 (deleted)'
            pass

        if not name:
            try:
                with open(os.path.join(pid_dir, 'stat'), 'r',
                          encoding='utf-8') as f:
                    name = f.read().split()[1].strip('()')
            except (OSError, IndexError):
                continue
        same_user = False
        try:
            same_user = os.stat(pid_dir).st_uid == target_uid
        except OSError:
            pass
        yield ProcessInfo(pid, name, same_user)


def _enumerate_ps_aux():
    """ps aux strategy (BSD/macOS)"""
    from bleachbit.General import get_real_username
    current_user = get_real_username()
    ps_out = subprocess.check_output(
        ["ps", "aux", "-c"], universal_newlines=True)
    first_line = ps_out.split('\n', maxsplit=1)[0].strip()
    if "USER" not in first_line or "COMMAND" not in first_line:
        raise RuntimeError("Unexpected ps header format")
    for line in ps_out.split("\n")[1:]:
        parts = line.split()
        if len(parts) < 11:
            continue
        yield ProcessInfo(int(parts[1]), parts[10], parts[0] == current_user)


def is_process_running(exename, require_same_user):
    """Check whether exename is running"""
    ci = os.name == 'nt'  # case-insensitive on Windows
    if ci:
        exename = exename.lower()
    for proc in enumerate_processes():
        name = proc.name.lower() if ci else proc.name
        if name == exename and (not require_same_user or proc.same_user):
            return True
    return False


def terminate_process(exename, require_same_user):
    """Terminate processes matching exename. Returns list of affected PIDs."""
    ci = os.name == 'nt'
    if ci:
        exename = exename.lower()
    terminated = []
    for proc in enumerate_processes():
        name = proc.name.lower() if ci else proc.name
        if name == exename and (not require_same_user or proc.same_user):
            if proc.pid in terminated:
                continue
            try:
                if IS_WINDOWS:
                    psutil.Process(proc.pid).kill()
                else:
                    os.kill(proc.pid, signal.SIGTERM)
                terminated.append(proc.pid)
            except (ProcessLookupError, PermissionError, OSError):
                continue
    return terminated
